from mpi4py import MPI
import mpiunittest as unittest
import arrayimpl
import sys

pypy_lt_53 = (hasattr(sys, 'pypy_version_info') and
              sys.pypy_version_info < (5, 3))

def mkzeros(n):
    if pypy_lt_53:
        return b'\0' * n
    return bytearray(n)

def memzero(m):
    try:
        m[:] = 0
    except IndexError: # cffi buffer
        m[0:len(m)] = b'\0'*len(m)

class BaseTestRMA(object):

    COMM = MPI.COMM_NULL
    INFO = MPI.INFO_NULL

    def setUp(self):
        nbytes = 100*MPI.DOUBLE.size
        try:
            self.mpi_memory = MPI.Alloc_mem(nbytes)
            self.memory = self.mpi_memory
            memzero(self.memory)
        except MPI.Exception:
            import array
            self.mpi_memory = None
            self.memory = array.array('B',[0]*nbytes)
        self.WIN = MPI.Win.Create(self.memory, 1, self.INFO, self.COMM)

    def tearDown(self):
        self.WIN.Free()
        if self.mpi_memory:
            MPI.Free_mem(self.mpi_memory)

    def testPutGet(self):
        typemap = MPI._typedict
        group = self.WIN.Get_group()
        size = group.Get_size()
        group.Free()
        for array, typecode in arrayimpl.subTest(self):
            if unittest.is_mpi_gpu('mvapich2', array): continue
            for count in range(10):
                for rank in range(size):
                    sbuf = array(range(count), typecode)
                    rbuf = array(-1, typecode, count+1)
                    #
                    self.WIN.Fence()
                    self.WIN.Put(sbuf.as_mpi(), rank)
                    self.WIN.Fence()
                    self.WIN.Get(rbuf.as_mpi_c(count), rank)
                    self.WIN.Fence()
                    for i in range(count):
                        self.assertEqual(sbuf[i], i)
                        self.assertNotEqual(rbuf[i], -1)
                    self.assertEqual(rbuf[-1], -1)
                    #
                    sbuf = array(range(count), typecode)
                    rbuf = array(-1, typecode, count+1)
                    target  = sbuf.itemsize
                    self.WIN.Fence()
                    self.WIN.Put(sbuf.as_mpi(), rank, target)
                    self.WIN.Fence()
                    self.WIN.Get(rbuf.as_mpi_c(count), rank, target)
                    self.WIN.Fence()
                    for i in range(count):
                        self.assertEqual(sbuf[i], i)
                        self.assertNotEqual(rbuf[i], -1)
                    self.assertEqual(rbuf[-1], -1)
                    #
                    sbuf = array(range(count), typecode)
                    rbuf = array(-1, typecode, count+1)
                    datatype = typemap[typecode]
                    target  = (sbuf.itemsize, count, datatype)
                    self.WIN.Fence()
                    self.WIN.Put(sbuf.as_mpi(), rank, target)
                    self.WIN.Fence()
                    self.WIN.Get(rbuf.as_mpi_c(count), rank, target)
                    self.WIN.Fence()
                    for i in range(count):
                        self.assertEqual(sbuf[i], i)
                        self.assertNotEqual(rbuf[i], -1)
                    self.assertEqual(rbuf[-1], -1)

    def testAccumulate(self):
        group = self.WIN.Get_group()
        size = group.Get_size()
        group.Free()
        for array, typecode in arrayimpl.subTest(self):
            if unittest.is_mpi_gpu('openmpi', array): continue
            if unittest.is_mpi_gpu('mvapich2', array): continue
            if typecode in 'FDG': continue
            for count in range(10):
                for rank in range(size):
                    sbuf = array(range(count), typecode)
                    rbuf = array(-1, typecode, count+1)
                    for op in (MPI.SUM, MPI.PROD, MPI.MAX, MPI.MIN):
                        self.WIN.Fence()
                        self.WIN.Accumulate(sbuf.as_mpi(), rank, op=op)
                        self.WIN.Fence()
                        self.WIN.Get(rbuf.as_mpi_c(count), rank)
                        self.WIN.Fence()
                        for i in range(count):
                            self.assertEqual(sbuf[i], i)
                            self.assertNotEqual(rbuf[i], -1)
                        self.assertEqual(rbuf[-1], -1)

    @unittest.skipMPI('openmpi(>=1.10,<1.11)')
    def testGetAccumulate(self):
        group = self.WIN.Get_group()
        size = group.Get_size()
        rank = group.Get_rank()
        group.Free()
        self.WIN.Fence()
        obuf = MPI.Alloc_mem(1); memzero(obuf)
        rbuf = MPI.Alloc_mem(1); memzero(rbuf)
        try:
            try:
                self.WIN.Get_accumulate([obuf, 0, MPI.BYTE], [rbuf, 0, MPI.BYTE], rank)
            finally:
                MPI.Free_mem(obuf)
                MPI.Free_mem(rbuf)
        except NotImplementedError:
            self.skipTest('mpi-win-get_accumulate')
        self.WIN.Fence()
        for array, typecode in arrayimpl.subTest(self):
            if unittest.is_mpi_gpu('openmpi', array): continue
            if unittest.is_mpi_gpu('mvapich2', array): continue
            if typecode in 'FDG': continue
            for count in range(10):
                for rank in range(size):
                    ones = array([1]*count, typecode)
                    sbuf = array(range(count), typecode)
                    rbuf = array(-1, typecode, count+1)
                    gbuf = array(-1, typecode, count+1)
                    for op in (MPI.SUM, MPI.PROD,
                               MPI.MAX, MPI.MIN,
                               MPI.REPLACE, MPI.NO_OP):
                        self.WIN.Lock(rank)
                        self.WIN.Put(ones.as_mpi(), rank)
                        self.WIN.Flush(rank)
                        self.WIN.Get_accumulate(sbuf.as_mpi(),
                                                rbuf.as_mpi_c(count),
                                                rank, op=op)
                        self.WIN.Flush(rank)
                        self.WIN.Get(gbuf.as_mpi_c(count), rank)
                        self.WIN.Flush(rank)
                        self.WIN.Unlock(rank)
                        #
                        for i in range(count):
                            self.assertEqual(sbuf[i], i)
                            self.assertEqual(rbuf[i], 1)
                            self.assertEqual(gbuf[i], op(1, i))
                        self.assertEqual(rbuf[-1], -1)
                        self.assertEqual(gbuf[-1], -1)

    def testFetchAndOp(self):
        typemap = MPI._typedict
        group = self.WIN.Get_group()
        size = group.Get_size()
        rank = group.Get_rank()
        group.Free()
        self.WIN.Fence()
        blen = MPI.INT.Get_size()
        obuf = MPI.Alloc_mem(blen); memzero(obuf)
        rbuf = MPI.Alloc_mem(blen); memzero(rbuf)
        try:
            try:
                self.WIN.Fetch_and_op(
                    [obuf, 1, MPI.INT],
                    [rbuf, 1, MPI.INT],
                    rank)
            finally:
                MPI.Free_mem(obuf)
                MPI.Free_mem(rbuf)
        except NotImplementedError:
            self.skipTest('mpi-win-fetch_and_op')
        self.WIN.Fence()
        for array, typecode in arrayimpl.subTest(self):
            if unittest.is_mpi_gpu('openmpi', array): continue
            if unittest.is_mpi_gpu('mvapich2', array): continue
            if typecode in 'FDG': continue
            obuf = array(+1, typecode)
            rbuf = array(-1, typecode, 2)
            datatype = typemap[typecode]
            for op in (MPI.SUM, MPI.PROD,
                       MPI.MAX, MPI.MIN,
                       MPI.REPLACE, MPI.NO_OP):
                for rank in range(size):
                    for disp in range(3):
                        self.WIN.Lock(rank)
                        self.WIN.Fetch_and_op(obuf.as_mpi(),
                                              rbuf.as_mpi_c(1),
                                              rank,
                                              disp * datatype.size,
                                              op=op)

                        self.WIN.Unlock(rank)
                        self.assertEqual(rbuf[1], -1)

    def testCompareAndSwap(self):
        typemap = MPI._typedict
        group = self.WIN.Get_group()
        size = group.Get_size()
        rank = group.Get_rank()
        group.Free()
        self.WIN.Fence()
        obuf = MPI.Alloc_mem(1); memzero(obuf)
        cbuf = MPI.Alloc_mem(1); memzero(cbuf)
        rbuf = MPI.Alloc_mem(1); memzero(rbuf)
        try:
            try:
                self.WIN.Compare_and_swap([obuf, 1, MPI.BYTE],
                                          [cbuf, 1, MPI.BYTE],
                                          [rbuf, 1, MPI.BYTE],
                                          rank, 0)
            finally:
                MPI.Free_mem(obuf)
                MPI.Free_mem(cbuf)
                MPI.Free_mem(rbuf)
        except NotImplementedError:
            self.skipTest('mpi-win-compare_and_swap')
        self.WIN.Fence()
        for array, typecode in arrayimpl.subTest(self):
            if unittest.is_mpi_gpu('openmpi', array): continue
            if unittest.is_mpi_gpu('mvapich2', array): continue
            if typecode in 'fdg': continue
            if typecode in 'FDG': continue
            obuf = array(+1, typecode)
            cbuf = array( 0, typecode)
            rbuf = array(-1, typecode, 2)
            datatype = typemap[typecode]
            for rank in range(size):
                for disp in range(3):
                    self.WIN.Lock(rank)
                    self.WIN.Compare_and_swap(obuf.as_mpi(),
                                              cbuf.as_mpi(),
                                              rbuf.as_mpi_c(1),
                                              rank,
                                              disp * datatype.size)

                    self.WIN.Unlock(rank)
                    self.assertEqual(rbuf[1], -1)

    def testPutProcNull(self):
        self.WIN.Fence()
        self.WIN.Put(None, MPI.PROC_NULL, None)
        self.WIN.Fence()

    def testGetProcNull(self):
        self.WIN.Fence()
        self.WIN.Get(None, MPI.PROC_NULL, None)
        self.WIN.Fence()

    def testAccumulateProcNullReplace(self):
        self.WIN.Fence()
        zeros = mkzeros(8)
        self.WIN.Fence()
        self.WIN.Accumulate([zeros, MPI.INT], MPI.PROC_NULL, None, MPI.REPLACE)
        self.WIN.Fence()
        self.WIN.Accumulate([zeros, MPI.INT], MPI.PROC_NULL, None, MPI.REPLACE)
        self.WIN.Fence()

    def testAccumulateProcNullSum(self):
        self.WIN.Fence()
        zeros = mkzeros(8)
        self.WIN.Fence()
        self.WIN.Accumulate([zeros, MPI.INT], MPI.PROC_NULL, None, MPI.SUM)
        self.WIN.Fence()
        self.WIN.Accumulate([None, MPI.INT], MPI.PROC_NULL, None, MPI.SUM)
        self.WIN.Fence()

    def testGetAccumulateProcNull(self):
        obuf = [mkzeros(8), 0, MPI.INT]
        rbuf = [mkzeros(8), 0, MPI.INT]
        self.WIN.Fence()
        try:
            self.WIN.Get_accumulate(obuf, rbuf, MPI.PROC_NULL)
        except NotImplementedError:
            self.skipTest('mpi-win-get_accumulate')
        self.WIN.Fence()

    ##def testFetchAndOpProcNull(self):
    ##    obuf = cbuf = rbuf = None
    ##    self.WIN.Fence()
    ##    try:
    ##        self.WIN.Fetch_and_op(obuf, rbuf, MPI.PROC_NULL, 0)
    ##    except NotImplementedError:
    ##        self.skipTest('mpi-win-fetch_and_op')
    ##    self.WIN.Fence()

    ##def testCompareAndSwapProcNull(self):
    ##    obuf = cbuf = rbuf = None
    ##    self.WIN.Fence()
    ##    try:
    ##        self.WIN.Compare_and_swap(obuf, cbuf, rbuf, MPI.PROC_NULL, 0)
    ##    except NotImplementedError:
    ##        self.skipTest('mpi-win-compare_and_swap')
    ##    self.WIN.Fence()

    def testFence(self):
        win = self.WIN
        LMODE = [0, MPI.MODE_NOSTORE, MPI.MODE_NOPUT,
                 MPI.MODE_NOSTORE|MPI.MODE_NOPUT]
        GMODE = [0, MPI.MODE_NOPRECEDE, MPI.MODE_NOSUCCEED]
        win.Fence()
        for lmode in LMODE:
            for gmode in GMODE:
                assertion =  lmode | gmode
                win.Fence(assertion)
        win.Fence()

    @unittest.skipMPI('openmpi(==1.8.1)')
    def testFenceAll(self):
        win = self.WIN
        assertion = 0
        modes = [0,
                 MPI.MODE_NOSTORE,
                 MPI.MODE_NOPUT,
                 MPI.MODE_NOPRECEDE,
                 MPI.MODE_NOSUCCEED]
        win.Fence()
        for mode in modes:
            win.Fence(mode)
            assertion |= mode
            win.Fence(assertion)
        win.Fence()

    @unittest.skipMPI('openmpi(==1.8.6)')
    def testStartComplete(self):
        self.WIN.Start(MPI.GROUP_EMPTY)
        self.WIN.Complete()

    @unittest.skipMPI('openmpi(==1.8.6)')
    def testPostWait(self):
        self.WIN.Post(MPI.GROUP_EMPTY)
        self.WIN.Wait()

    @unittest.skipMPI('openmpi(==1.8.7)')
    @unittest.skipMPI('openmpi(==1.8.6)')
    def testStartCompletePostWait(self):
        win = self.WIN
        wingroup = win.Get_group()
        size = wingroup.Get_size()
        rank = wingroup.Get_rank()
        if size < 2: return wingroup.Free()
        if rank == 0:
            group = wingroup.Excl([0])
            win.Start(group)
            win.Complete()
            win.Post(group)
            win.Wait()
            group.Free()
        else:
            group = wingroup.Incl([0])
            win.Post(group)
            win.Wait()
            win.Start(group)
            win.Complete()
            group.Free()
        wingroup.Free()

    @unittest.skipMPI('openmpi(==1.8.7)')
    @unittest.skipMPI('openmpi(==1.8.6)')
    def testStartCompletePostTest(self):
        comm = self.COMM
        win = self.WIN
        wingroup = win.Get_group()
        size = wingroup.Get_size()
        rank = wingroup.Get_rank()
        if size < 2: return wingroup.Free()
        if rank == 0:
            group = wingroup.Excl([0])
            win.Start(group)
            comm.Barrier()
            win.Complete()
            comm.Barrier()
            group.Free()
        else:
            group = wingroup.Incl([0])
            win.Post(group)
            flag = win.Test()
            self.assertFalse(flag)
            comm.Barrier()
            comm.Barrier()
            flag = win.Test()
            self.assertTrue(flag)
            group.Free()
        wingroup.Free()

    @unittest.skipMPI('MPI(<3.0)')
    def testSync(self):
        win = self.WIN
        comm = self.COMM
        rank = comm.Get_rank()
        win.Lock(rank)
        win.Sync()
        win.Unlock(rank)
        comm.Barrier()

    @unittest.skipMPI('MPI(<3.0)')
    def testFlush(self):
        win = self.WIN
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        #
        for i in range(size):
            win.Lock(i)
            win.Flush(i)
            win.Unlock(i)
        comm.Barrier()
        for i in range(size):
            if i == rank:
                win.Lock_all()
                win.Flush_all()
                win.Unlock_all()
            comm.Barrier()
        #
        for i in range(size):
            win.Lock(i)
            win.Flush_local(i)
            win.Unlock(i)
        comm.Barrier()
        for i in range(size):
            if i == rank:
                win.Lock_all()
                win.Flush_local_all()
                win.Unlock_all()
            comm.Barrier()

class TestRMASelf(BaseTestRMA, unittest.TestCase):
    COMM = MPI.COMM_SELF

class TestRMAWorld(BaseTestRMA, unittest.TestCase):
    COMM = MPI.COMM_WORLD


SpectrumMPI = MPI.get_vendor()[0] == 'Spectrum MPI'
try:
    if SpectrumMPI: raise NotImplementedError
    MPI.Win.Create(None, 1, MPI.INFO_NULL, MPI.COMM_SELF).Free()
except (NotImplementedError, MPI.Exception):
    unittest.disable(BaseTestRMA, 'mpi-rma')


if __name__ == '__main__':
    unittest.main()



# -----------------------------------------------------------------------------

cdef extern from "atimport.h": pass

# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    enum: PY3 "(PY_MAJOR_VERSION>=3)"
    enum: PY2 "(PY_MAJOR_VERSION==2)"
    enum: PYPY "PyMPI_RUNTIME_PYPY"

    void PySys_WriteStderr(char*,...)
    int Py_AtExit(void (*)())

    ctypedef struct PyObject
    PyObject *Py_None
    void Py_CLEAR(PyObject*)

    void Py_INCREF(object)
    void Py_DECREF(object)

# -----------------------------------------------------------------------------

cdef extern from *:
    const char *Py_GETENV(const char[])
    enum: USE_MATCHED_RECV "PyMPI_USE_MATCHED_RECV"

ctypedef struct Options:
    int initialize
    int threads
    int thread_level
    int finalize
    int fast_reduce
    int recv_mprobe
    int errors

cdef Options options
options.initialize = 1
options.threads = 1
options.thread_level = MPI_THREAD_MULTIPLE
options.finalize = 1
options.fast_reduce = 1
options.recv_mprobe = 1
options.errors = 1

cdef object getEnv(object rc, const char name[], object value):
    cdef bytes oname = b"MPI4PY_RC_" + name.upper()
    cdef const char *cvalue = Py_GETENV(oname)
    if cvalue == NULL: return value
    cdef object ovalue = pystr(cvalue)
    cdef bytes  bvalue = PyBytes_FromString(cvalue).lower()
    if bvalue in (b'true',  b'yes', b'on',  b'y', b'1'): ovalue = True
    if bvalue in (b'false', b'no',  b'off', b'n', b'0'): ovalue = False
    try: setattr(rc, pystr(name), ovalue)
    except: pass
    return ovalue

cdef int warnOpt(object name, object value) except -1:
    cdef object warn
    from warnings import warn
    warn("mpi4py.rc: '%s': unexpected value %r" % (name, value))

cdef int getOptions(Options* opts) except -1:
    cdef object rc
    opts.initialize = 1
    opts.threads = 1
    opts.thread_level = MPI_THREAD_MULTIPLE
    opts.finalize = 1
    opts.fast_reduce = 1
    opts.recv_mprobe = 1
    opts.errors = 1
    try: from mpi4py import rc
    except: return 0
    #
    cdef object initialize = True
    cdef object threads = True
    cdef object thread_level = 'multiple'
    cdef object finalize = None
    cdef object fast_reduce = True
    cdef object recv_mprobe = True
    cdef object errors = 'exception'
    try: initialize = rc.initialize
    except: pass
    try: threads = rc.threads
    except: pass
    try: threads = rc.threaded # backward
    except: pass               # compatibility
    try: thread_level = rc.thread_level
    except: pass
    try: finalize = rc.finalize
    except: pass
    try: fast_reduce = rc.fast_reduce
    except: pass
    try: recv_mprobe = rc.recv_mprobe
    except: pass
    try: errors = rc.errors
    except: pass
    initialize   = getEnv(rc, b"initialize",   initialize)
    threads      = getEnv(rc, b"threads",      threads)
    thread_level = getEnv(rc, b"thread_level", thread_level)
    finalize     = getEnv(rc, b"finalize",     finalize)
    fast_reduce  = getEnv(rc, b"fast_reduce",  fast_reduce)
    recv_mprobe  = getEnv(rc, b"recv_mprobe",  recv_mprobe)
    errors       = getEnv(rc, b"errors",       errors)
    #
    if initialize in (True, 'yes'):
        opts.initialize = 1
    elif initialize in (False, 'no'):
        opts.initialize = 0
    else:
        warnOpt("initialize", initialize)
    #
    if threads in (True, 'yes'):
        opts.threads = 1
    elif threads in (False, 'no'):
        opts.threads = 0
    else:
        warnOpt("threads", threads)
    #
    if thread_level == 'single':
        opts.thread_level = MPI_THREAD_SINGLE
    elif thread_level == 'funneled':
        opts.thread_level = MPI_THREAD_FUNNELED
    elif thread_level == 'serialized':
        opts.thread_level = MPI_THREAD_SERIALIZED
    elif thread_level == 'multiple':
        opts.thread_level = MPI_THREAD_MULTIPLE
    else:
        warnOpt("thread_level", thread_level)
    #
    if finalize is None:
        opts.finalize = opts.initialize
    elif finalize in (True, 'yes'):
        opts.finalize = 1
    elif finalize in (False, 'no'):
        opts.finalize = 0
    else:
        warnOpt("finalize", finalize)
    #
    if fast_reduce in (True, 'yes'):
        opts.fast_reduce = 1
    elif fast_reduce in (False, 'no'):
        opts.fast_reduce = 0
    else:
        warnOpt("fast_reduce", fast_reduce)
    #
    if recv_mprobe in (True, 'yes'):
        opts.recv_mprobe = 1 and USE_MATCHED_RECV
    elif recv_mprobe in (False, 'no'):
        opts.recv_mprobe = 0
    else:
        warnOpt("recv_mprobe", recv_mprobe)
    #
    if errors == 'default':
        opts.errors = 0
    elif errors == 'exception':
        opts.errors = 1
    elif errors == 'fatal':
        opts.errors = 2
    else:
        warnOpt("errors", errors)
    #
    return 0

# -----------------------------------------------------------------------------

cdef extern from *:
    int PyMPI_Commctx_finalize() nogil

cdef int bootstrap() except -1:
    # Get options from 'mpi4py.rc' module
    getOptions(&options)
    # Cleanup at (the very end of) Python exit
    if Py_AtExit(atexit) < 0:
        PySys_WriteStderr(b"warning: could not register "
                          b"cleanup with Py_AtExit()%s", b"\n")
    # Do we have to initialize MPI?
    cdef int initialized = 1
    <void>MPI_Initialized(&initialized)
    if initialized:
        options.finalize = 0
        return 0
    if not options.initialize:
        return 0
    # MPI initialization
    cdef int ierr = MPI_SUCCESS
    cdef int required = MPI_THREAD_SINGLE
    cdef int provided = MPI_THREAD_SINGLE
    if options.threads:
        required = options.thread_level
        ierr = MPI_Init_thread(NULL, NULL, required, &provided)
        if ierr != MPI_SUCCESS: raise RuntimeError(
            "MPI_Init_thread() failed [error code: %d]" % ierr)
    else:
        ierr = MPI_Init(NULL, NULL)
        if ierr != MPI_SUCCESS: raise RuntimeError(
            "MPI_Init() failed [error code: %d]" % ierr)
    return 0

cdef inline int mpi_active() nogil:
    cdef int ierr = MPI_SUCCESS
    # MPI initialized ?
    cdef int initialized = 0
    ierr = MPI_Initialized(&initialized)
    if not initialized or ierr != MPI_SUCCESS: return 0
    # MPI finalized ?
    cdef int finalized = 1
    ierr = MPI_Finalized(&finalized)
    if finalized or ierr != MPI_SUCCESS: return 0
    # MPI should be active ...
    return 1

cdef int initialize() nogil except -1:
    if not mpi_active(): return 0
    comm_set_eh(MPI_COMM_SELF)
    comm_set_eh(MPI_COMM_WORLD)
    return 0

cdef void finalize() nogil:
    if not mpi_active(): return
    <void>PyMPI_Commctx_finalize()

cdef int abort_status = 0

cdef void atexit() nogil:
    if not mpi_active(): return
    if abort_status:
        <void>MPI_Abort(MPI_COMM_WORLD, abort_status)
    finalize()
    if options.finalize:
        <void>MPI_Finalize()

def _set_abort_status(object status: Any) -> None:
    "Helper for ``python -m mpi4py.run ...``"
    global abort_status
    try:
        abort_status = status
    except:
        abort_status = 1 if status else 0

# -----------------------------------------------------------------------------

# Vile hack for raising a exception and not contaminate the traceback

cdef extern from *:
    enum: PyMPI_ERR_UNAVAILABLE

cdef extern from "Python.h":
    void PyErr_SetObject(object, object)
    void *PyExc_RuntimeError
    void *PyExc_NotImplementedError

cdef object MPIException = <object>PyExc_RuntimeError

cdef int PyMPI_Raise(int ierr) except -1 with gil:
    if ierr == PyMPI_ERR_UNAVAILABLE:
        PyErr_SetObject(<object>PyExc_NotImplementedError, None)
        return 0
    if (<void*>MPIException) != NULL:
        PyErr_SetObject(MPIException, <long>ierr)
    else:
        PyErr_SetObject(<object>PyExc_RuntimeError, <long>ierr)
    return 0

cdef inline int CHKERR(int ierr) nogil except -1:
    if ierr == MPI_SUCCESS: return 0
    PyMPI_Raise(ierr)
    return -1

cdef inline void print_traceback():
    cdef object sys, traceback
    import sys, traceback
    traceback.print_exc()
    try: sys.stderr.flush()
    except: pass

# -----------------------------------------------------------------------------

# PyPy: Py_IsInitialized() cannot be called without the GIL

cdef extern from "Python.h":
    int _Py_IsInitialized"Py_IsInitialized"() nogil

cdef object _pypy_sentinel = None

cdef inline int Py_IsInitialized() nogil:
    if PYPY and (<void*>_pypy_sentinel) == NULL: return 0
    return _Py_IsInitialized()

# -----------------------------------------------------------------------------



class Exception(RuntimeError):

    """
    Exception class
    """

    def __init__(self, int ierr: int = 0):
        if ierr < MPI_SUCCESS: ierr = MPI_ERR_UNKNOWN
        self.ob_mpi = ierr
        RuntimeError.__init__(self, self.ob_mpi)

    def __eq__(self, object error) -> bool:
        cdef int ierr = self.ob_mpi
        return <bint> (ierr == error)

    def __ne__(self, object error) -> bool:
        cdef int ierr = self.ob_mpi
        return <bint> (ierr != error)

    def __lt__(self, object error) -> bool:
        cdef int ierr = self.ob_mpi
        return <bint> (ierr < error)

    def __le__(self, object error) -> bool:
        cdef int ierr = self.ob_mpi
        return <bint> (ierr <= error)

    def __gt__(self, object error) -> bool:
        cdef int ierr = self.ob_mpi
        return <bint> (ierr > error)

    def __ge__(self, object error) -> bool:
        cdef int ierr = self.ob_mpi
        return <bint> (ierr >= error)

    def __hash__(self) -> int:
        return hash(self.ob_mpi)

    def __bool__(self) -> bool:
        cdef int ierr = self.ob_mpi
        return ierr != MPI_SUCCESS

    def __int__(self) -> int:
        return self.ob_mpi

    def __repr__(self) -> str:
        return "MPI.Exception(%d)" % self.ob_mpi

    def __str__(self) -> str:
        if not mpi_active():
            return "error code: %d" % self.ob_mpi
        return self.Get_error_string()

    def Get_error_code(self) -> int:
        """
        Error code
        """
        cdef int errorcode = MPI_SUCCESS
        errorcode = self.ob_mpi
        return errorcode

    error_code = property(Get_error_code, doc="error code")

    def Get_error_class(self) -> int:
        """
        Error class
        """
        cdef int errorclass = MPI_SUCCESS
        CHKERR( MPI_Error_class(self.ob_mpi, &errorclass) )
        return errorclass

    error_class = property(Get_error_class, doc="error class")

    def Get_error_string(self) -> str:
        """
        Error string
        """
        cdef char string[MPI_MAX_ERROR_STRING+1]
        cdef int resultlen = 0
        CHKERR( MPI_Error_string(self.ob_mpi, string, &resultlen) )
        return tompistr(string, resultlen)

    error_string = property(Get_error_string, doc="error string")

    if PY2: __nonzero__ = __bool__



cdef extern from "mpi-compat.h": pass

cimport mpi4py.MPI as MPI
from mpi4py.libmpi cimport *

cdef extern from "stdio.h":
    int printf(char*, ...)

cdef void c_sayhello(MPI_Comm comm):
    cdef int size, rank, plen
    cdef char pname[MPI_MAX_PROCESSOR_NAME]
    if comm == MPI_COMM_NULL:
        printf(b"You passed MPI_COMM_NULL !!!%s", b"\n")
        return
    MPI_Comm_size(comm, &size)
    MPI_Comm_rank(comm, &rank)
    MPI_Get_processor_name(pname, &plen)
    printf(b"Hello, World! I am process %d of %d on %s.\n",
           rank, size, pname)

def sayhello(MPI.Comm comm not None ):
    cdef MPI_Comm c_comm = comm.ob_mpi
    c_sayhello(c_comm)



import sys; sys.path.insert(0, sys.argv[1])
import mpi4py
if len(sys.argv) > 2:
    lfn = "runtests-mpi4py-child"
    mpe = sys.argv[2] == 'mpe'
    vt  = sys.argv[2] == 'vt'
    if mpe: mpi4py.profile('mpe', logfile=lfn)
    if vt:  mpi4py.profile('vt',  logfile=lfn)
from mpi4py import MPI
parent = MPI.Comm.Get_parent()
parent.Barrier()
parent.Disconnect()
assert parent == MPI.COMM_NULL
parent = MPI.Comm.Get_parent()
assert parent == MPI.COMM_NULL



#------------------------------------------------------------------------------

# CUDA array interface for interoperating Python CUDA GPU libraries
# See https://numba.pydata.org/numba-doc/latest/cuda/cuda_array_interface.html

cdef inline int cuda_is_contig(tuple shape,
                               tuple strides,
                               Py_ssize_t itemsize,
                               char order) except -1:
    cdef Py_ssize_t i, ndim = len(shape)
    cdef Py_ssize_t start, step, index, dim, size = itemsize
    if order == c'F':
        start = 0
        step = 1
    else:
        start = ndim - 1
        step = -1
    for i from 0 <= i < ndim:
        index = start + step * i
        dim = <Py_ssize_t>shape[index]
        if dim > 1 and size != <Py_ssize_t>strides[index]:
            return 0
        size *= dim
    return 1

cdef inline char* cuda_get_format(char typekind, Py_ssize_t itemsize) nogil:
   if typekind == c'b':
       if itemsize == sizeof(char): return b"?"
   if typekind == c'i':
       if itemsize == sizeof(char):      return b"b"
       if itemsize == sizeof(short):     return b"h"
       if itemsize == sizeof(int):       return b"i"
       if itemsize == sizeof(long):      return b"l"
       if itemsize == sizeof(long long): return b"q"
   if typekind == c'u':
       if itemsize == sizeof(char):      return b"B"
       if itemsize == sizeof(short):     return b"H"
       if itemsize == sizeof(int):       return b"I"
       if itemsize == sizeof(long):      return b"L"
       if itemsize == sizeof(long long): return b"Q"
   if typekind == c'f':
       if itemsize == sizeof(float)//2:    return b"e"
       if itemsize == sizeof(float):       return b"f"
       if itemsize == sizeof(double):      return b"d"
       if itemsize == sizeof(long double): return b"g"
   if typekind == c'c':
       if itemsize == 2*sizeof(float)//2:    return b"Ze"
       if itemsize == 2*sizeof(float):       return b"Zf"
       if itemsize == 2*sizeof(double):      return b"Zd"
       if itemsize == 2*sizeof(long double): return b"Zg"
   return BYTE_FMT

#------------------------------------------------------------------------------

cdef int Py_CheckCAIBuffer(object obj):
    try: return <bint>hasattr(obj, '__cuda_array_interface__')
    except: return 0

cdef int Py_GetCAIBuffer(object obj, Py_buffer *view, int flags) except -1:
    cdef dict cuda_array_interface
    cdef tuple data
    cdef str   typestr
    cdef tuple shape
    cdef tuple strides
    cdef list descr
    cdef object dev_ptr, mask
    cdef void *buf = NULL
    cdef bint readonly = 0
    cdef Py_ssize_t s, size = 1
    cdef Py_ssize_t itemsize = 1
    cdef char typekind = c'u'
    cdef bint fixnull = 0

    try:
        cuda_array_interface = obj.__cuda_array_interface__
    except AttributeError:
        raise NotImplementedError("missing CUDA array interface")

    # mandatory
    data = cuda_array_interface['data']
    typestr = cuda_array_interface['typestr']
    shape = cuda_array_interface['shape']

    # optional
    strides = cuda_array_interface.get('strides')
    descr = cuda_array_interface.get('descr')
    mask = cuda_array_interface.get('mask')

    dev_ptr, readonly = data
    for s in shape: size *= s
    if dev_ptr is None and size == 0: dev_ptr = 0 # XXX
    buf = PyLong_AsVoidPtr(dev_ptr)
    typekind = <char>ord(typestr[1])
    itemsize = <Py_ssize_t>int(typestr[2:])

    if mask is not None:
        raise BufferError(
            "__cuda_array_interface__: "
            "cannot handle masked arrays"
        )
    if size < 0:
        raise BufferError(
            "__cuda_array_interface__: "
            "buffer with negative size (shape:%s, size:%d)"
            % (shape, size)
        )
    if (strides is not None and
        not cuda_is_contig(shape, strides, itemsize, c'C') and
        not cuda_is_contig(shape, strides, itemsize, c'F')):
        raise BufferError(
            "__cuda_array_interface__: "
            "buffer is not contiguous (shape:%s, strides:%s, itemsize:%d)"
            % (shape, strides, itemsize)
        )
    if descr is not None and (len(descr) != 1 or descr[0] != ('', typestr)):
        PyErr_WarnEx(RuntimeWarning,
                     b"__cuda_array_interface__: "
                     b"ignoring 'descr' key", 1)

    if PYPY and readonly and ((flags & PyBUF_WRITABLE) == PyBUF_WRITABLE):
        raise BufferError("Object is not writable")

    fixnull = (buf == NULL and size == 0)
    if fixnull: buf = &fixnull
    PyBuffer_FillInfo(view, obj, buf, size*itemsize, readonly, flags)
    if fixnull: view.buf = NULL

    if (flags & PyBUF_FORMAT) == PyBUF_FORMAT:
        view.format = cuda_get_format(typekind, itemsize)
        if view.format != BYTE_FMT:
            view.itemsize = itemsize
    return 0

#------------------------------------------------------------------------------



from mpi4py import MPI
import mpiunittest as unittest
import sys

datatypes_c = [
MPI.CHAR, MPI.WCHAR,
MPI.SIGNED_CHAR, MPI.SHORT, MPI.INT, MPI.LONG,
MPI.UNSIGNED_CHAR, MPI.UNSIGNED_SHORT, MPI.UNSIGNED, MPI.UNSIGNED_LONG,
MPI.LONG_LONG, MPI.UNSIGNED_LONG_LONG,
MPI.FLOAT, MPI.DOUBLE, MPI.LONG_DOUBLE,
]
datatypes_c99 = [
MPI.C_BOOL,
MPI.INT8_T, MPI.INT16_T, MPI.INT32_T, MPI.INT64_T,
MPI.UINT8_T, MPI.UINT16_T, MPI.UINT32_T, MPI.UINT64_T,
MPI.C_COMPLEX, MPI.C_FLOAT_COMPLEX,
MPI.C_DOUBLE_COMPLEX, MPI.C_LONG_DOUBLE_COMPLEX,
]
datatypes_f = [
MPI.CHARACTER, MPI.LOGICAL, MPI.INTEGER,
MPI.REAL, MPI.DOUBLE_PRECISION,
MPI.COMPLEX, MPI.DOUBLE_COMPLEX,
]
datatypes_f90 = [
MPI.LOGICAL1, MPI.LOGICAL2, MPI.LOGICAL4, MPI.LOGICAL8,
MPI.INTEGER1, MPI.INTEGER2, MPI.INTEGER4, MPI.INTEGER8, MPI.INTEGER16,
MPI.REAL2, MPI.REAL4, MPI.REAL8, MPI.REAL16,
MPI.COMPLEX4, MPI.COMPLEX8, MPI.COMPLEX16, MPI.COMPLEX32,
]
datatypes_mpi = [
MPI.PACKED, MPI.BYTE, MPI.AINT, MPI.OFFSET,
]

datatypes = []
datatypes += datatypes_c
datatypes += datatypes_c99
datatypes += datatypes_f
datatypes += datatypes_f90
datatypes += datatypes_mpi
datatypes = [t for t in datatypes if t != MPI.DATATYPE_NULL]

combiner_map = {}

class TestDatatype(unittest.TestCase):

    def testBoolEqNe(self):
        for dtype in datatypes:
            self.assertTrue (not not dtype)
            self.assertTrue (dtype == MPI.Datatype(dtype))
            self.assertFalse(dtype != MPI.Datatype(dtype))

    def testGetExtent(self):
        for dtype in datatypes:
            lb, ext = dtype.Get_extent()
            self.assertEqual(dtype.lb, lb)
            self.assertEqual(dtype.ub, lb+ext)
            self.assertEqual(dtype.extent, ext)

    def testGetSize(self):
        for dtype in datatypes:
            size = dtype.Get_size()
            self.assertTrue(dtype.size, size)

    def testGetTrueExtent(self):
        for dtype in datatypes:
            try:
                lb, ext = dtype.Get_true_extent()
                self.assertEqual(dtype.true_lb, lb)
                self.assertEqual(dtype.true_ub, lb+ext)
                self.assertEqual(dtype.true_extent, ext)
            except NotImplementedError:
                self.skipTest('mpi-type-get_true_extent')

    def testGetEnvelope(self):
        for dtype in datatypes:
            try:
                envelope = dtype.Get_envelope()
            except NotImplementedError:
                self.skipTest('mpi-type-get_envelope')
            if ('LAM/MPI' == MPI.get_vendor()[0] and
                "COMPLEX" in dtype.name): continue
            ni, na, nd, combiner = envelope
            self.assertEqual(combiner, MPI.COMBINER_NAMED)
            self.assertEqual(ni, 0)
            self.assertEqual(na, 0)
            self.assertEqual(nd, 0)
            self.assertEqual(dtype.envelope, envelope)
            self.assertEqual(dtype.combiner, combiner)
            self.assertTrue(dtype.is_named)
            self.assertTrue(dtype.is_predefined)
            otype = dtype.decode()
            self.assertTrue(dtype is otype)

    def check_datatype_contents(self, oldtype, factory, newtype):
        try:
            envelope = newtype.Get_envelope()
            contents = newtype.Get_contents()
        except NotImplementedError:
            self.skipTest('mpi-type-get_envelope')
        ni, na, nd, combiner = envelope
        i, a, d = contents
        self.assertEqual(ni, len(i))
        self.assertEqual(na, len(a))
        self.assertEqual(nd, len(d))
        self.assertTrue(combiner != MPI.COMBINER_NAMED)
        self.assertEqual(newtype.envelope, envelope)
        self.assertEqual(newtype.contents, contents)
        self.assertEqual(newtype.combiner, combiner)
        self.assertFalse(newtype.is_named)
        if combiner in (MPI.COMBINER_F90_INTEGER,
                        MPI.COMBINER_F90_REAL,
                        MPI.COMBINER_F90_COMPLEX,):
            self.assertTrue(newtype.is_predefined)
        else:
            self.assertFalse(newtype.is_predefined)
        name = factory.__name__
        NAME = name.replace('Create_', '').upper()
        symbol = getattr(MPI, 'COMBINER_' + NAME)
        if symbol == MPI.UNDEFINED: return
        if combiner_map is None: return
        symbol = combiner_map.get(symbol, symbol)
        if symbol is None: return
        self.assertEqual(symbol, combiner)
        decoded = newtype.decode()
        oldtype, constructor, kargs = decoded
        constructor = 'Create_' + constructor.lower()
        newtype2 = getattr(oldtype, constructor)(**kargs)
        decoded2 = newtype2.decode()
        self.assertEqual(decoded[1], decoded2[1])
        self.assertEqual(decoded[2], decoded2[2])
        if combiner not in (MPI.COMBINER_F90_INTEGER,
                            MPI.COMBINER_F90_REAL,
                            MPI.COMBINER_F90_COMPLEX,):
            self.assertFalse(newtype2.is_predefined)
            newtype2.Free()
        else:
            self.assertTrue(newtype2.is_predefined)

    def check_datatype(self, oldtype, factory, *args):
        try:
            if isinstance(oldtype, MPI.Datatype):
                newtype = factory(oldtype, *args)
            else:
                newtype = factory(*args)
        except NotImplementedError:
            self.skipTest('mpi-type-constructor')
        self.check_datatype_contents(oldtype, factory,  newtype)
        newtype.Commit()
        self.check_datatype_contents(oldtype, factory,  newtype)
        combiner = newtype.Get_envelope()[-1]
        if combiner not in (MPI.COMBINER_F90_INTEGER,
                            MPI.COMBINER_F90_REAL,
                            MPI.COMBINER_F90_COMPLEX,):
            newtype.Free()

    def testDup(self):
        for dtype in datatypes:
            factory = MPI.Datatype.Dup
            self.check_datatype(dtype, factory)

    def testCreateContiguous(self):
        for dtype in datatypes:
            for count in range(5):
                factory = MPI.Datatype.Create_contiguous
                args = (count, )
                self.check_datatype(dtype, factory, *args)

    def testCreateVector(self):
        for dtype in datatypes:
            for count in range(5):
                for blocklength in range(5):
                    for stride in range(5):
                        factory = MPI.Datatype.Create_vector
                        args = (count, blocklength, stride)
                        self.check_datatype(dtype, factory, *args)

    def testCreateHvector(self):
        for dtype in datatypes:
            for count in range(5):
                for blocklength in range(5):
                    for stride in range(5):
                        factory = MPI.Datatype.Create_hvector
                        args = (count, blocklength, stride)
                        self.check_datatype(dtype, factory, *args)

    def testCreateIndexed(self):
        for dtype in datatypes:
            for block in range(5):
                blocklengths = list(range(block, block+5))
                displacements = [0]
                for b in blocklengths[:-1]:
                    stride = displacements[-1] + b * dtype.extent + 1
                    displacements.append(stride)
                factory = MPI.Datatype.Create_indexed
                args = (blocklengths, displacements)
                self.check_datatype(dtype, factory, *args)
                #args = (block, displacements) XXX
                #self.check_datatype(dtype, factory, *args)  XXX

    def testCreateIndexedBlock(self):
        for dtype in datatypes:
            for block in range(5):
                blocklengths = list(range(block, block+5))
                displacements = [0]
                for b in blocklengths[:-1]:
                    stride = displacements[-1] + b * dtype.extent + 1
                    displacements.append(stride)
                factory = MPI.Datatype.Create_indexed_block
                args = (block, displacements)
                self.check_datatype(dtype, factory, *args)

    def testCreateHindexed(self):
        for dtype in datatypes:
            for block in range(5):
                blocklengths = list(range(block, block+5))
                displacements = [0]
                for b in blocklengths[:-1]:
                    stride = displacements[-1] + b * dtype.extent + 1
                    displacements.append(stride)

                factory = MPI.Datatype.Create_hindexed
                args = (blocklengths, displacements)
                self.check_datatype(dtype, factory, *args)
                #args = (block, displacements) XXX
                #self.check_datatype(dtype, factory, *args)  XXX

    @unittest.skipMPI('openmpi(<=1.8.1)', MPI.VERSION == 3)
    def testCreateHindexedBlock(self):
        for dtype in datatypes:
            for block in range(5):
                displacements = [0]
                for i in range(5):
                    stride = displacements[-1] + block * dtype.extent + 1
                    displacements.append(stride)
                factory = MPI.Datatype.Create_hindexed_block
                args = (block, displacements)
                self.check_datatype(dtype, factory, *args)

    def testCreateStruct(self):
        for dtype1 in datatypes:
            for dtype2 in datatypes:
                dtypes = (dtype1, dtype2)
                blocklengths  = (2, 3)
                displacements = [0]
                for dtype in dtypes[:-1]:
                    stride = displacements[-1] + dtype.extent
                    displacements.append(stride)
                factory = MPI.Datatype.Create_struct
                args = (blocklengths, displacements, dtypes)
                self.check_datatype(dtypes, factory, *args)

    def testCreateSubarray(self):
        for dtype in datatypes:
            for ndim in range(1, 5):
                for size in range(1, 5):
                    for subsize in range(1, size):
                        for start in range(size-subsize):
                            for order in [MPI.ORDER_C,
                                          MPI.ORDER_FORTRAN,
                                          MPI.ORDER_F,
                                          ]:
                                sizes = [size] * ndim
                                subsizes = [subsize] * ndim
                                starts = [start] * ndim
                                factory = MPI.Datatype.Create_subarray
                                args = sizes, subsizes, starts, order
                                self.check_datatype(dtype, factory, *args)

    def testCreateDarray(self):
        for dtype in datatypes:
            for ndim in range(1, 3+1):
                for size in (4, 8, 9, 27):
                    for rank in (0, size-1):
                        for dist in [MPI.DISTRIBUTE_BLOCK, MPI.DISTRIBUTE_CYCLIC]:
                            for order in [MPI.ORDER_C, MPI.ORDER_F]:
                                gsizes = [size]*ndim
                                distribs = [dist]*ndim
                                dargs = [MPI.DISTRIBUTE_DFLT_DARG]*ndim
                                psizes = MPI.Compute_dims(size, [0]*ndim)
                                factory = MPI.Datatype.Create_darray
                                args = size, rank, gsizes, distribs, dargs, psizes, order
                                self.check_datatype(dtype, factory, *args)

    def testCreateF90Integer(self):
        for r in (1, 2, 4):
            factory = MPI.Datatype.Create_f90_integer
            args = (r,)
            self.check_datatype(None, factory, *args)

    @unittest.skipMPI('openmpi(<3.0.0)')
    @unittest.skipMPI('msmpi')
    @unittest.skipMPI('SpectrumMPI')
    def testCreateF90RealSingle(self):
        (p, r) = (6, 30)
        factory = MPI.Datatype.Create_f90_real
        args = (p, r)
        self.check_datatype(None, factory, *args)

    @unittest.skipMPI('openmpi(<3.0.0)')
    @unittest.skipMPI('msmpi')
    @unittest.skipMPI('SpectrumMPI')
    def testCreateF90RealDouble(self):
        (p, r) = (15, 300)
        factory = MPI.Datatype.Create_f90_real
        args = (p, r)
        self.check_datatype(None, factory, *args)

    @unittest.skipMPI('openmpi(<3.0.0)')
    @unittest.skipMPI('msmpi')
    @unittest.skipMPI('SpectrumMPI')
    def testCreateF90ComplexSingle(self):
        (p, r) = (6, 30)
        factory = MPI.Datatype.Create_f90_complex
        args = (p, r)
        self.check_datatype(None, factory, *args)

    @unittest.skipMPI('openmpi(<3.0.0)')
    @unittest.skipMPI('msmpi')
    @unittest.skipMPI('SpectrumMPI')
    def testCreateF90ComplexDouble(self):
        (p, r) = (15, 300)
        factory = MPI.Datatype.Create_f90_complex
        args = (p, r)
        self.check_datatype(None, factory, *args)

    match_size_integer = [1, 2, 4, 8]
    match_size_real    = [4, 8]
    match_size_complex = [8, 16]
    @unittest.skipMPI('MPI(<2.0)')
    @unittest.skipMPI('openmpi', (MPI.CHARACTER == MPI.DATATYPE_NULL or
                                  MPI.CHARACTER.Get_size() == 0))
    def testMatchSize(self):
        typeclass = MPI.TYPECLASS_INTEGER
        for size in self.match_size_integer:
            datatype = MPI.Datatype.Match_size(typeclass, size)
            self.assertEqual(size, datatype.size)
        typeclass = MPI.TYPECLASS_REAL
        for size in self.match_size_real:
            datatype = MPI.Datatype.Match_size(typeclass, size)
            self.assertEqual(size, datatype.size)
        typeclass  = MPI.TYPECLASS_COMPLEX
        for size in self.match_size_complex:
            datatype = MPI.Datatype.Match_size(typeclass, size)
            self.assertEqual(size, datatype.size)

    def testCreateResized(self):
        for dtype in datatypes:
            for lb in range(-10, 10):
                for extent in range(1, 10):
                    factory = MPI.Datatype.Create_resized
                    args = lb, extent
                    self.check_datatype(dtype, factory, *args)

    def testGetSetName(self):
        for dtype in datatypes:
            try:
                name = dtype.Get_name()
                self.assertTrue(name)
                dtype.Set_name(name)
                self.assertEqual(name, dtype.Get_name())
            except NotImplementedError:
                self.skipTest('mpi-type-name')

    def testCommit(self):
        for dtype in datatypes:
            dtype.Commit()


name, version = MPI.get_vendor()
if name == 'LAM/MPI':
    combiner_map[MPI.COMBINER_INDEXED_BLOCK] = MPI.COMBINER_INDEXED
elif name == 'MPICH1':
    combiner_map[MPI.COMBINER_VECTOR]  = None
    combiner_map[MPI.COMBINER_HVECTOR] = None
    combiner_map[MPI.COMBINER_INDEXED] = None
    combiner_map[MPI.COMBINER_HINDEXED_BLOCK] = None
    for t in datatypes_f: datatypes.remove(t)
elif MPI.Get_version() < (2,0):
    combiner_map = None
if name == 'Open MPI':
    for t in datatypes_f + datatypes_f90:
        if t != MPI.DATATYPE_NULL:
            if t.Get_size() == 0:
                if t in datatypes:
                    datatypes.remove(t)
    if (1,6,0) < version < (1,7,0):
        TestDatatype.match_size_complex[:] = []
    if version < (1,5,2):
        for t in datatypes_f90[-4:]:
            if t != MPI.DATATYPE_NULL:
                datatypes.remove(t)
if name == 'Platform MPI':
    combiner_map[MPI.COMBINER_INDEXED_BLOCK] = MPI.COMBINER_INDEXED
    combiner_map[MPI.COMBINER_DARRAY] = MPI.COMBINER_STRUCT
    combiner_map[MPI.COMBINER_SUBARRAY] = MPI.COMBINER_STRUCT
    TestDatatype.match_size_complex[:] = []


if __name__ == '__main__':
    unittest.main()



cdef class Message:

    """
    Matched message handle
    """

    def __cinit__(self, Message message: Optional[Message] = None):
        self.ob_mpi = MPI_MESSAGE_NULL
        if message is None: return
        self.ob_mpi = message.ob_mpi
        self.ob_buf = message.ob_buf

    def __dealloc__(self):
        if not (self.flags & PyMPI_OWNED): return
        CHKERR( del_Message(&self.ob_mpi) )

    def __richcmp__(self, other, int op):
        if not isinstance(other, Message): return NotImplemented
        cdef Message s = <Message>self, o = <Message>other
        if   op == Py_EQ: return (s.ob_mpi == o.ob_mpi)
        elif op == Py_NE: return (s.ob_mpi != o.ob_mpi)
        cdef mod = type(self).__module__
        cdef cls = type(self).__name__
        raise TypeError("unorderable type: '%s.%s'" % (mod, cls))

    def __bool__(self) -> bool:
        return self.ob_mpi != MPI_MESSAGE_NULL

    # Matching Probe
    # --------------

    @classmethod
    def Probe(
        cls,
        Comm comm: Comm,
        int source: int = ANY_SOURCE,
        int tag: int = ANY_TAG,
        Status status: Optional[Status] = None,
    ) -> Message:
        """
        Blocking test for a matched message
        """
        cdef MPI_Message cmessage = MPI_MESSAGE_NULL
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_Mprobe(
            source, tag, comm.ob_mpi, &cmessage, statusp) )
        cdef Message message = <Message>Message.__new__(cls)
        message.ob_mpi = cmessage
        return message

    @classmethod
    def Iprobe(
        cls,
        Comm comm: Comm,
        int source: int = ANY_SOURCE,
        int tag: int = ANY_TAG,
        Status status: Optional[Status] = None,
    ) -> Optional[Message]:
        """
        Nonblocking test for a matched message
        """
        cdef int flag = 0
        cdef MPI_Message cmessage = MPI_MESSAGE_NULL
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_Improbe(
             source, tag, comm.ob_mpi, &flag, &cmessage, statusp) )
        if flag == 0: return None
        cdef Message message = <Message>Message.__new__(cls)
        message.ob_mpi = cmessage
        return message

    # Matched receives
    # ----------------

    def Recv(
        self,
        buf: BufSpec,
        Status status: Optional[Status] = None,
    ) -> None:
        """
        Blocking receive of matched message
        """
        cdef MPI_Message message = self.ob_mpi
        cdef int source = MPI_ANY_SOURCE
        if message == MPI_MESSAGE_NO_PROC:
            source = MPI_PROC_NULL
        cdef _p_msg_p2p rmsg = message_p2p_recv(buf, source)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_Mrecv(
            rmsg.buf, rmsg.count, rmsg.dtype,
            &message, statusp) )
        if self is not __MESSAGE_NO_PROC__:
            self.ob_mpi = message

    def Irecv(self, buf: BufSpec) -> Request:
        """
        Nonblocking receive of matched message
        """
        cdef MPI_Message message = self.ob_mpi
        cdef int source = MPI_ANY_SOURCE
        if message == MPI_MESSAGE_NO_PROC:
            source = MPI_PROC_NULL
        cdef _p_msg_p2p rmsg = message_p2p_recv(buf, source)
        cdef Request request = Request.__new__(Request)
        with nogil: CHKERR( MPI_Imrecv(
            rmsg.buf, rmsg.count, rmsg.dtype,
            &message, &request.ob_mpi) )
        if self is not __MESSAGE_NO_PROC__:
            self.ob_mpi = message
        request.ob_buf = rmsg
        return request

    # Python Communication
    # --------------------
    #
    @classmethod
    def probe(
        cls,
        Comm comm: Comm,
        int source: int = ANY_SOURCE,
        int tag: int = ANY_TAG,
        Status status: Optional[Status] = None,
    ) -> Message:
        """Blocking test for a matched message"""
        cdef Message message = <Message>Message.__new__(cls)
        cdef MPI_Status *statusp = arg_Status(status)
        message.ob_buf = PyMPI_mprobe(source, tag, comm.ob_mpi,
                                      &message.ob_mpi, statusp)
        return message
    #
    @classmethod
    def iprobe(
        cls, Comm comm: Comm,
        int source: int = ANY_SOURCE,
        int tag: int = ANY_TAG,
        Status status: Optional[Status] = None,
    ) -> Optional[Message]:
        """Nonblocking test for a matched message"""
        cdef int flag = 0
        cdef Message message = <Message>Message.__new__(cls)
        cdef MPI_Status *statusp = arg_Status(status)
        message.ob_buf = PyMPI_improbe(source, tag, comm.ob_mpi, &flag,
                                       &message.ob_mpi, statusp)
        if flag == 0: return None
        return message
    #
    def recv(self, Status status: Optional[Status] = None) -> Any:
        """Blocking receive of matched message"""
        cdef object rmsg = self.ob_buf
        cdef MPI_Message message = self.ob_mpi
        cdef MPI_Status *statusp = arg_Status(status)
        rmsg = PyMPI_mrecv(rmsg, &message, statusp)
        if self is not __MESSAGE_NO_PROC__: self.ob_mpi = message
        if self.ob_mpi == MPI_MESSAGE_NULL: self.ob_buf = None
        return rmsg
    #
    def irecv(self) -> Request:
        """Nonblocking receive of matched message"""
        cdef object rmsg = self.ob_buf
        cdef MPI_Message message = self.ob_mpi
        cdef Request request = Request.__new__(Request)
        request.ob_buf = PyMPI_imrecv(rmsg, &message, &request.ob_mpi)
        if self is not __MESSAGE_NO_PROC__: self.ob_mpi = message
        if self.ob_mpi == MPI_MESSAGE_NULL: self.ob_buf = None
        return request

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Message_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Message:
        """
        """
        cdef Message message = Message.__new__(Message)
        message.ob_mpi = MPI_Message_f2c(arg)
        return message


cdef Message __MESSAGE_NULL__    = new_Message ( MPI_MESSAGE_NULL    )
cdef Message __MESSAGE_NO_PROC__ = new_Message ( MPI_MESSAGE_NO_PROC )


# Predefined message handles
# --------------------------

MESSAGE_NULL    = __MESSAGE_NULL__    #: Null message handle
MESSAGE_NO_PROC = __MESSAGE_NO_PROC__ #: No-proc message handle



#!/usr/bin/env python

# If you want MPE to log MPI calls, you have to add the two lines
# below at the very beginning of your main bootstrap script.
import mpi4py
mpi4py.profile('mpe', logfile='cpilog')

# Import the MPI extension module
from mpi4py import MPI
if 0: # <- use '1' to disable logging of MPI calls
    MPI.Pcontrol(0)

# Import the 'array' module
from array import array

# This is just to make the logging
# output a bit more interesting
from time import sleep

comm = MPI.COMM_WORLD
nprocs = comm.Get_size()
myrank = comm.Get_rank()

n  = array('i', [0])
pi = array('d', [0])
mypi = array('d', [0])

def comp_pi(n, myrank=0, nprocs=1):
    h = 1.0 / n;
    s = 0.0;
    for i in range(myrank + 1, n + 1, nprocs):
        x = h * (i - 0.5);
        s += 4.0 / (1.0 + x**2);
    return s * h

comm.Barrier()

for N in [10000]*10:

    if myrank == 0:
        n[0] = N

    comm.Bcast([n, MPI.INT], root=0)

    mypi[0] = comp_pi(n[0], myrank, nprocs)

    comm.Reduce([mypi, MPI.DOUBLE],
                [pi, MPI.DOUBLE],
                op=MPI.SUM, root=0)

    comm.Barrier()

    sleep(0.01)



from mpi4py import MPI
import sys, os, optparse

assert __name__ == '__main__'
from os.path import split, splitext, dirname, realpath
dirname = dirname(__file__)
assert sys.path[0] == realpath(dirname)
if split(__file__)[1] == '__main__.py':
    if splitext(dirname)[0] == '.zip':
        assert sys.argv[0] == dirname
    else:
        assert realpath(sys.argv[0]) == realpath(dirname)
else:
    assert sys.argv[0] == __file__

parser = optparse.OptionParser()
parser.add_option("--rank", action='store',
                  type='int', dest="rank", default=0)
parser.add_option("--sys-exit", action='store',
                  type='int', dest="sys_exit", default=None)
parser.add_option("--sys-exit-msg", action="store",
                  type="string", dest="sys_exit", default=None)
parser.add_option("--exception", action="store",
                  type="string", dest="exception", default=None)
(options, args) = parser.parse_args()
assert not args

comm = MPI.COMM_WORLD
if comm.rank == options.rank:
    if options.sys_exit:
        sys.exit(options.sys_exit)
    if options.exception:
        raise RuntimeError(options.exception)

comm.Barrier()
if comm.rank > 0:
    comm.Recv([None, 'B'], comm.rank - 1)
print("Hello, World!")
if comm.rank < comm.size - 1:
    comm.Send([None, 'B'], comm.rank + 1)
comm.Barrier()

sys.exit()



from mpi4py import MPI
import mpiunittest as unittest

class BaseTestFortran(object):
    HANDLES = []
    def testFortran(self):
        for handle1 in self.HANDLES:
            try:
                fint = handle1.py2f()
            except NotImplementedError:
                continue
            handle2 = type(handle1).f2py(fint)
            self.assertEqual(handle1, handle2)

class TestFortranStatus(BaseTestFortran, unittest.TestCase):
    def setUp(self):
        s1 = MPI.Status()
        s2 = MPI.Status()
        s2.source = 1
        s2.tag = 2
        s2.error = MPI.ERR_OTHER
        s3 = MPI.Status()
        s3.source = 0
        s3.tag = 0
        s3.error = MPI.SUCCESS
        self.HANDLES = [s1, s2, s3]

    @unittest.skipMPI('MPICH1')
    def testFortran(self):
        super(TestFortranStatus, self).testFortran()


class TestFortranDatatype(BaseTestFortran, unittest.TestCase):
    HANDLES = [MPI.DATATYPE_NULL,
               MPI.CHAR,  MPI.SHORT,
               MPI.INT,   MPI.LONG,
               MPI.FLOAT, MPI.DOUBLE,
               ]

class TestFortranOp(BaseTestFortran, unittest.TestCase):
    HANDLES =  [MPI.OP_NULL,
                MPI.MAX, MPI.MIN,
                MPI.SUM, MPI.PROD,
                MPI.LAND, MPI.BAND,
                MPI.LOR, MPI.BOR,
                MPI.LXOR, MPI.BXOR,
                MPI.MAXLOC, MPI.MINLOC,
                ]

class TestFortranRequest(BaseTestFortran, unittest.TestCase):
    HANDLES = [MPI.REQUEST_NULL,
               ]

class TestFortranMessage(BaseTestFortran, unittest.TestCase):
    HANDLES = [MPI.MESSAGE_NULL,
               MPI.MESSAGE_NO_PROC,
               ]

class TestFortranErrhandler(BaseTestFortran, unittest.TestCase):
    HANDLES = [MPI.ERRHANDLER_NULL,
               MPI.ERRORS_RETURN,
               MPI.ERRORS_ARE_FATAL,
               ]

class TestFortranInfo(BaseTestFortran, unittest.TestCase):
    HANDLES = [MPI.INFO_NULL,
               ]

class TestFortranGroup(BaseTestFortran, unittest.TestCase):
    HANDLES = [MPI.GROUP_NULL,
               MPI.GROUP_EMPTY,
               ]

class TestFortranComm(BaseTestFortran, unittest.TestCase):
    HANDLES = [MPI.COMM_NULL,
               MPI.COMM_SELF,
               MPI.COMM_WORLD,
               ]

class TestFortranWin(BaseTestFortran, unittest.TestCase):
    HANDLES = [MPI.WIN_NULL,
               ]

class TestFortranFile(BaseTestFortran, unittest.TestCase):
    HANDLES = [MPI.FILE_NULL,
               ]


if __name__ == '__main__':
    unittest.main()



from mpi4py import rc
rc.initialize = False

from mpi4py import MPI
assert not MPI.Is_initialized()
assert not MPI.Is_finalized()

MPI.Init_thread(MPI.THREAD_MULTIPLE)
assert MPI.Is_initialized()
assert not MPI.Is_finalized()

MPI.Finalize()
assert MPI.Is_initialized()
assert MPI.Is_finalized()



#!/usr/bin/env python
# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com

import os
import argparse
import requests

APIURL  = 'https://ci.appveyor.com/api'
ACCOUNT = 'mpi4py/mpi4py'
BRANCH  = 'maint'

parser = argparse.ArgumentParser(prog=os.path.basename(__file__))
parser.add_argument("-q", "--quiet", action="store_false",
                    dest="verbose", default=True)
parser.add_argument("-n", "--dry-run", action="store_false",
                    dest="download", default=True)
parser.add_argument("-a", "--account",  type=str, action="store",
                    dest="account", default=ACCOUNT)
parser.add_argument("-b", "--branch",  type=str, action="store",
                    dest="branch", default=BRANCH)
options = parser.parse_args()

ACCOUNT = options.account
BRANCH  = options.branch

branch_url = APIURL + '/projects/' + ACCOUNT + "/branch/" + BRANCH
branch = requests.get(branch_url).json()
jobs = branch['build']['jobs']
jobids = [job['jobId'] for job in jobs]

if options.verbose:
    print("Downloading AppVeyor artifacts "
          "account={} branch={}".format(ACCOUNT, BRANCH))
for jobid in jobids:
    artifacts_url = APIURL + '/buildjobs/' + jobid + '/artifacts'
    artifacts = requests.get(artifacts_url).json()
    filenames = [a['fileName'] for a in artifacts]
    for filename in filenames:
        download_url = artifacts_url + '/' + filename
        if options.verbose:
            print(download_url)
        if options.download:
            data = requests.get(download_url).content
            with open(os.path.basename(filename), "wb") as f:
                f.write(data)



# ---

import mpi4py
try: mpi4py.get_include()
except: pass
try: mpi4py.get_config()
except: pass

# ---

def test_mpi4py_rc():
    import mpi4py.rc
    mpi4py.rc(
    initialize = True,
    threads = True,
    thread_level = 'multiple',
    finalize = None,
    fast_reduce = True,
    recv_mprobe = True,
    errors = 'exception',
    )
    try: mpi4py.rc(qwerty=False)
    except TypeError: pass
    else: raise RuntimeError
    repr(mpi4py.rc)

test_mpi4py_rc()

# ---

def test_mpi4py_profile():
    import mpi4py
    def mpi4py_profile(*args, **kargs):
        try: mpi4py.profile(*args, **kargs)
        except ValueError: pass
    import warnings
    warnings.simplefilter('ignore')
    mpi4py_profile('mpe')
    mpi4py_profile('mpe', path="/usr/lib")
    mpi4py_profile('mpe', path=["/usr/lib"])
    mpi4py_profile('mpe', logfile="mpi4py")
    mpi4py_profile('mpe', logfile="mpi4py")
    mpi4py_profile('vt')
    mpi4py_profile('vt', path="/usr/lib")
    mpi4py_profile('vt', path=["/usr/lib"])
    mpi4py_profile('vt', logfile="mpi4py")
    mpi4py_profile('vt', logfile="mpi4py")
    mpi4py_profile('@querty')
    mpi4py_profile('c', path=["/usr/lib", "/usr/lib64"])
    mpi4py_profile('m', path=["/usr/lib", "/usr/lib64"])
    mpi4py_profile('dl', path=["/usr/lib", "/usr/lib64"])
    mpi4py_profile('hosts', path=["/etc"])

test_mpi4py_profile()

# ---

import mpi4py.__main__
import mpi4py.bench
import mpi4py.futures
import mpi4py.futures.__main__
import mpi4py.futures.server
import mpi4py.run



from mpi4py import MPI
import mpiunittest as unittest

_basic = [None,
          True, False,
          -7, 0, 7,
          -2**63+1, 2**63-1,
          -2.17, 0.0, 3.14,
          1+2j, 2-3j,
          'mpi4py',
          ]
messages = list(_basic)
messages += [ list(_basic),
              tuple(_basic),
              dict([('k%d' % key, val)
                    for key, val in enumerate(_basic)])
              ]

@unittest.skipIf(MPI.MESSAGE_NULL == MPI.MESSAGE_NO_PROC, 'mpi-message')
class TestMessage(unittest.TestCase):

    def testMessageNull(self):
        null = MPI.MESSAGE_NULL
        self.assertFalse(null)
        null2 = MPI.Message()
        self.assertEqual(null, null2)
        null3 = MPI.Message(null)
        self.assertEqual(null, null3)

    def testMessageNoProc(self):
        #
        noproc = MPI.MESSAGE_NO_PROC
        self.assertTrue(noproc)
        noproc.recv()
        self.assertTrue(noproc)
        noproc.irecv().wait()
        self.assertTrue(noproc)
        #
        noproc2 = MPI.Message(MPI.MESSAGE_NO_PROC)
        self.assertTrue(noproc2)
        self.assertEqual(noproc2, noproc)
        self.assertNotEqual(noproc, MPI.MESSAGE_NULL)
        #
        message = MPI.Message(MPI.MESSAGE_NO_PROC)
        message.recv()
        self.assertEqual(message, MPI.MESSAGE_NULL)
        #
        message = MPI.Message(MPI.MESSAGE_NO_PROC)
        request = message.irecv()
        self.assertEqual(message, MPI.MESSAGE_NULL)
        self.assertNotEqual(request, MPI.REQUEST_NULL)
        request.wait()
        self.assertEqual(request, MPI.REQUEST_NULL)
        #
        comm = MPI.COMM_SELF
        message = comm.mprobe(MPI.PROC_NULL)
        self.assertNotEqual(message, MPI.MESSAGE_NULL)
        self.assertEqual(message, MPI.MESSAGE_NO_PROC)
        noproc = comm.improbe(MPI.PROC_NULL)
        self.assertNotEqual(message, MPI.MESSAGE_NULL)
        self.assertEqual(message, MPI.MESSAGE_NO_PROC)

@unittest.skipIf(MPI.MESSAGE_NULL == MPI.MESSAGE_NO_PROC, 'mpi-message')
class BaseTestP2PMatched(object):

    COMM = MPI.COMM_NULL

    def testIMProbe(self):
        comm = self.COMM.Dup()
        try:
            m = comm.improbe()
            self.assertEqual(m, None)
            m = comm.improbe(MPI.ANY_SOURCE)
            self.assertEqual(m, None)
            m = comm.improbe(MPI.ANY_SOURCE, MPI.ANY_TAG)
            self.assertEqual(m, None)
            status = MPI.Status()
            m = comm.improbe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertEqual(m, None)
            self.assertEqual(status.source, MPI.ANY_SOURCE)
            self.assertEqual(status.tag,    MPI.ANY_TAG)
            self.assertEqual(status.error,  MPI.SUCCESS)
            m = MPI.Message.iprobe(comm)
            self.assertEqual(m, None)
            s = comm.isend(None, comm.rank, 0)
            r = comm.mprobe(comm.rank, 0).irecv()
            MPI.Request.waitall([s,r])
        finally:
            comm.Free()

    def testProbeRecv(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        for smsg in messages:
            if size == 1:
                sr = comm.isend(smsg, 0, 0)
                m = comm.mprobe(0, 0)
                self.assertTrue(isinstance(m, MPI.Message))
                self.assertTrue(m)
                rr = m.irecv()
                self.assertFalse(m)
                self.assertTrue(sr)
                self.assertTrue(rr)
                MPI.Request.Waitall([sr,rr])
                self.assertFalse(sr)
                self.assertFalse(rr)
                #
                r = comm.isend(smsg, 0, 0)
                m = MPI.Message.probe(comm, 0, 0)
                self.assertTrue(isinstance(m, MPI.Message))
                self.assertTrue(m)
                rmsg = m.recv()
                self.assertFalse(m)
                r.wait()
            elif rank == 0:
                comm.send(smsg, 1, 0)
                m = comm.mprobe(1, 0)
                self.assertTrue(m)
                rmsg = m.recv()
                self.assertFalse(m)
                #
                comm.send(smsg, 1, 1)
                m = None
                while not m:
                    m = MPI.Message.iprobe(comm, 1, 1)
                rmsg = m.irecv().wait()
            elif rank == 1:
                m = comm.mprobe(0, 0)
                self.assertTrue(m)
                rmsg = m.recv()
                self.assertFalse(m)
                comm.send(rmsg, 0, 0)
                #
                m = None
                while not m:
                    m = MPI.Message.iprobe(comm, 0, 1)
                rmsg = m.irecv().wait()
                comm.send(smsg, 0, 1)
            else:
                rmsg = smsg

            self.assertEqual(smsg, rmsg)


class TestP2PMatchedSelf(BaseTestP2PMatched, unittest.TestCase):
    COMM = MPI.COMM_SELF

class TestP2PMatchedWorld(BaseTestP2PMatched, unittest.TestCase):
    COMM = MPI.COMM_WORLD

@unittest.skipMPI('openmpi(<1.8.5)', MPI.COMM_WORLD.Get_size() > 1)
class TestP2PMatchedSelfDup(TestP2PMatchedSelf):
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()
    def tearDown(self):
        self.COMM.Free()

@unittest.skipMPI('openmpi(<1.8.5)', MPI.COMM_WORLD.Get_size() > 1)
class TestP2PMatchedWorldDup(TestP2PMatchedWorld):
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()
    def tearDown(self):
        self.COMM.Free()


if __name__ == '__main__':
    unittest.main()



from mpi4py import MPI
import mpiunittest as unittest

try:
    import array
except ImportError:
    array = None
try:
    import numpy
except ImportError:
    numpy = None

class TestAddress(unittest.TestCase):

    @unittest.skipIf(array is None, 'array')
    def testGetAddress1(self):
        from struct import pack, unpack
        location = array.array('i', range(10))
        bufptr, _ = location.buffer_info()
        addr = MPI.Get_address(location)
        addr = unpack('P', pack('P', addr))[0]
        self.assertEqual(addr, bufptr)

    @unittest.skipIf(numpy is None, 'numpy')
    def testGetAddress2(self):
        from struct import pack, unpack
        location = numpy.asarray(range(10), dtype='i')
        bufptr, _ = location.__array_interface__['data']
        addr = MPI.Get_address(location)
        addr = unpack('P', pack('P', addr))[0]
        self.assertEqual(addr, bufptr)

    def testBottom(self):
        base = MPI.Get_address(MPI.BOTTOM)
        addr = MPI.Aint_add(base, 0)
        self.assertEqual(addr, base)
        diff = MPI.Aint_diff(base, base)
        self.assertEqual(diff, 0)

    @unittest.skipIf(array is None, 'array')
    def testAintAdd(self):
        location = array.array('i', range(10))
        base = MPI.Get_address(location)
        addr = MPI.Aint_add(base, 4)
        self.assertEqual(addr, base + 4)

    @unittest.skipIf(array is None, 'array')
    def testAintDiff(self):
        location = array.array('i', range(10))
        base = MPI.Get_address(location)
        addr1 = base + 8
        addr2 = base + 4
        diff = MPI.Aint_diff(addr1, addr2)
        self.assertEqual(diff, 4)


if __name__ == '__main__':
    unittest.main()



# -----------------------------------------------------------------------------

ctypedef fused integral_t:
    int
    MPI_Aint

cdef inline object newarray(int n, integral_t **p):
    return allocate(n, sizeof(integral_t), p)

cdef inline object getarray(object ob, int *n, integral_t **p):
    cdef Py_ssize_t  olen = len(ob)
    cdef integral_t *base = NULL
    cdef int i = 0, size = downcast(olen)
    cdef object mem = newarray(size, &base)
    for i from 0 <= i < size: base[i] = ob[i]
    n[0] = size
    p[0] = base
    return mem

cdef inline object chkarray(object ob, int n, integral_t **p):
    cdef int size = 0
    cdef object mem = getarray(ob, &size, p)
    if n != size: raise ValueError(
        "expecting %d items, got %d" % (n, size))
    return mem

# -----------------------------------------------------------------------------

cdef inline object asarray_Datatype(object sequence,
                                    int size, MPI_Datatype **p):
     cdef int i = 0
     cdef MPI_Datatype *array = NULL
     if size != len(sequence): raise ValueError(
         "expecting %d items, got %d" % (size, len(sequence)))
     cdef object ob = allocate(size, sizeof(MPI_Datatype), &array)
     for i from 0 <= i < size:
         array[i] = (<Datatype?>sequence[i]).ob_mpi
     p[0] = array
     return ob

cdef inline object asarray_Info(object sequence,
                                int size, MPI_Info **p):
     cdef int i = 0
     cdef MPI_Info *array = NULL
     cdef MPI_Info info = MPI_INFO_NULL
     cdef object ob
     if sequence is None or isinstance(sequence, Info):
         if sequence is not None:
             info = (<Info?>sequence).ob_mpi
         ob = allocate(size, sizeof(MPI_Info), &array)
         for i from 0 <= i < size:
             array[i] = info
     else:
         if size != len(sequence): raise ValueError(
             "expecting %d items, got %d" % (size, len(sequence)))
         ob = allocate(size, sizeof(MPI_Datatype), &array)
         for i from 0 <= i < size:
             array[i] = (<Info?>sequence[i]).ob_mpi
     p[0] = array
     return ob

# -----------------------------------------------------------------------------

cdef inline int is_string(object obj):
     return (isinstance(obj, str) or
             isinstance(obj, bytes) or
             isinstance(obj, unicode))

cdef inline object asstring(object ob, char *s[]):
     cdef Py_ssize_t n = 0
     cdef char *p = NULL, *q = NULL
     ob = asmpistr(ob, &p)
     PyBytes_AsStringAndSize(ob, &p, &n)
     cdef object mem = allocate(n+1, sizeof(char), &q)
     <void>memcpy(q, p, <size_t>n)
     q[n] = 0; s[0] = q;
     return mem

cdef inline object asarray_str(object sequence, char ***p):
     cdef char** array = NULL
     cdef Py_ssize_t i = 0, size = len(sequence)
     cdef object ob = allocate(size+1, sizeof(char*), &array)
     for i from 0 <= i < size:
         sequence[i] = asstring(sequence[i], &array[i])
     array[size] = NULL
     p[0] = array
     return (sequence, ob)

cdef inline object asarray_argv(object sequence, char ***p):
     if sequence is None:
         p[0] = MPI_ARGV_NULL
         return None
     if is_string(sequence):
         sequence = [sequence]
     else:
         sequence = list(sequence)
     return asarray_str(sequence, p)

cdef inline object asarray_cmds(object sequence, int *count, char ***p):
     if is_string(sequence):
         raise ValueError("expecting a sequence of strings")
     sequence = list(sequence)
     count[0] = <int>len(sequence)
     return asarray_str(sequence, p)

cdef inline object asarray_argvs(object sequence, int size, char ****p):
     if sequence is None:
         p[0] = MPI_ARGVS_NULL
         return None
     if is_string(sequence):
         sequence = [sequence] * size
     else:
         sequence = list(sequence)
         if size != len(sequence): raise ValueError(
             "expecting %d items, got %d" % (size, len(sequence)))
     cdef int i = 0
     cdef char*** array = NULL
     cdef object ob = allocate(size+1, sizeof(char**), &array)
     cdef object argv
     for i from 0 <= i < size:
         argv = sequence[i]
         if argv is None: argv = []
         sequence[i] = asarray_argv(argv, &array[i])
     array[size] = NULL
     p[0] = array
     return (sequence, ob)

cdef inline object asarray_nprocs(object sequence, int size, int **p):
     cdef object ob
     cdef int *array = NULL
     cdef int i = 0, value = 1
     if sequence is None or is_integral(sequence):
         if sequence is not None:
             value = sequence
         ob = newarray(size, &array)
         for i from 0 <= i < size:
             array[i] = value
     else:
         ob = chkarray(sequence, size, &array)
     p[0] = array
     return ob

# -----------------------------------------------------------------------------



from mpi4py import MPI
from mpi4py.util import pkl5
import unittest

_basic = [
    None,
    True, False,
    -7, 0, 7,
    -2**63+1, 2**63-1,
    -2.17, 0.0, 3.14,
    1+2j, 2-3j,
    'mpi4py',
]
messages = list(_basic)
messages += [
    list(_basic),
    tuple(_basic),
    set(_basic),
    frozenset(_basic),
    dict(('k%d' % k, v) for k, v in enumerate(_basic)),
]

try:
    import numpy
    array1 = numpy.arange(1, dtype='i')
    array2 = numpy.arange(1, dtype='f')
    messages.append(array1)
    messages.append(array2)
    messages.append([array1, array1])
    messages.append([array2, array2])
    messages.append((array1, array2))
except ImportError:
    numpy = None


class BaseTest(object):

    COMM = MPI.COMM_NULL
    CommType = MPI.Intracomm
    MessageType = MPI.Message
    RequestType = MPI.Request

    def setUp(self):
        self.COMM = self.CommType(self.COMM)
        self.bigmpi_prev = pkl5._bigmpi
        self.bigmpi = pkl5._BigMPI()
        pkl5._bigmpi = self.bigmpi

    def tearDown(self):
        pkl5._bigmpi = self.bigmpi_prev

    def testSendAndRecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            self.COMM.send(smess, MPI.PROC_NULL)
            rmess = self.COMM.recv(None, MPI.PROC_NULL, 0)
            self.assertEqual(rmess, None)
        if size == 1: return
        for smess in messages:
            if rank == 0:
                self.COMM.send(smess, rank+1, 0)
                rmess = smess
            elif rank == size - 1:
                rmess = self.COMM.recv(None, rank-1, 0)
            else:
                rmess = self.COMM.recv(None, rank-1, 0)
                self.COMM.send(rmess, rank+1, 0)
            self.assertEqual(rmess, smess)

    def testISendAndRecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        buf = None
        for smess in messages:
            req = self.COMM.isend(smess, MPI.PROC_NULL)
            self.assertTrue(req)
            req.wait()
            self.assertFalse(req)
            rmess = self.COMM.recv(buf, MPI.PROC_NULL, 0)
            self.assertEqual(rmess, None)
        for smess in messages:
            req = self.COMM.isend(smess, rank, 0)
            self.assertTrue(req)
            rmess = self.COMM.recv(buf, rank, 0)
            self.assertTrue(req)
            flag, _ = req.test()
            self.assertTrue(flag)
            self.assertFalse(req)
            self.assertEqual(rmess, smess)
        for smess in messages:
            dst = (rank+1)%size
            src = (rank-1)%size
            req = self.COMM.isend(smess, dst, 0)
            self.assertTrue(req)
            rmess = self.COMM.recv(buf, src, 0)
            req.wait()
            self.assertFalse(req)
            self.assertEqual(rmess, smess)

    def testSSendAndRecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            self.COMM.ssend(smess, MPI.PROC_NULL)
            rmess = self.COMM.recv(None, MPI.PROC_NULL, 0)
            self.assertEqual(rmess, None)
        if size == 1: return
        for smess in messages:
            if rank == 0:
                self.COMM.ssend(smess, rank+1, 0)
                rmess = smess
            elif rank == size - 1:
                rmess = self.COMM.recv(None, rank-1, 0)
            else:
                rmess = self.COMM.recv(None, rank-1, 0)
                self.COMM.ssend(rmess, rank+1, 0)
            self.assertEqual(rmess, smess)

    def testISSendAndRecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            req = self.COMM.issend(smess, MPI.PROC_NULL)
            self.assertTrue(req)
            req.wait()
            self.assertFalse(req)
            rmess = self.COMM.recv(None, MPI.PROC_NULL, 0)
            self.assertEqual(rmess, None)
        for smess in messages:
            req = self.COMM.issend(smess, rank, 0)
            self.assertTrue(req)
            flag, _ = req.test()
            self.assertFalse(flag)
            self.assertTrue(req)
            rmess = self.COMM.recv(None, rank, 0)
            self.assertTrue(req)
            req.wait()
            self.assertFalse(req)
            self.assertEqual(rmess, smess)
        for smess in messages:
            dst = (rank+1)%size
            src = (rank-1)%size
            req = self.COMM.issend(smess, dst, 0)
            self.assertTrue(req)
            rmess = self.COMM.recv(None, src, 0)
            req.wait()
            self.assertFalse(req)
            self.assertEqual(rmess, smess)

    def testBSendAndRecv(self):
        buf = MPI.Alloc_mem((1<<16)+MPI.BSEND_OVERHEAD)
        MPI.Attach_buffer(buf)
        try:
            size = self.COMM.Get_size()
            rank = self.COMM.Get_rank()
            for smess in messages:
                self.COMM.bsend(smess, MPI.PROC_NULL)
                rmess = self.COMM.recv(None, MPI.PROC_NULL, 0)
                self.assertEqual(rmess, None)
            if size == 1: return
            for smess in messages:
                if rank == 0:
                    self.COMM.bsend(smess, rank+1, 0)
                    rmess = smess
                elif rank == size - 1:
                    rmess = self.COMM.recv(None, rank-1, 0)
                else:
                    rmess = self.COMM.recv(None, rank-1, 0)
                    self.COMM.bsend(rmess, rank+1, 0)
                self.assertEqual(rmess, smess)
        finally:
            MPI.Detach_buffer()
            MPI.Free_mem(buf)


    def testIBSendAndRecv(self):
        buf = MPI.Alloc_mem((1<<16)+MPI.BSEND_OVERHEAD)
        MPI.Attach_buffer(buf)
        try:
            size = self.COMM.Get_size()
            rank = self.COMM.Get_rank()
            for smess in messages:
                req = self.COMM.ibsend(smess, MPI.PROC_NULL)
                self.assertTrue(req)
                req.wait()
                self.assertFalse(req)
                rmess = self.COMM.recv(None, MPI.PROC_NULL, 0)
                self.assertEqual(rmess, None)
            for smess in messages:
                req = self.COMM.ibsend(smess, rank, 0)
                self.assertTrue(req)
                rmess = self.COMM.recv(None, rank, 0)
                self.assertTrue(req)
                flag, _ = req.test()
                self.assertTrue(flag)
                self.assertFalse(req)
                self.assertEqual(rmess, smess)
            for smess in messages:
                dst = (rank+1)%size
                src = (rank-1)%size
                req = self.COMM.ibsend(smess, dst, 0)
                self.assertTrue(req)
                rmess = self.COMM.recv(None, src, 0)
                req.wait()
                self.assertFalse(req)
                self.assertEqual(rmess, smess)
        finally:
            MPI.Detach_buffer()
            MPI.Free_mem(buf)

    def testISSendCancel(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            req = self.COMM.issend(smess, rank, 0)
            self.assertTrue(req)
            flag, _ = req.test()
            self.assertFalse(flag)
            self.assertTrue(req)
            req.cancel()
            self.assertTrue(req)
            self.assertTrue (req != MPI.REQUEST_NULL)
            self.assertFalse(req == MPI.REQUEST_NULL)
            status = MPI.Status()
            req.get_status(status)
            if not status.Is_cancelled():
                rmess = self.COMM.recv(None, rank, 0)
                self.assertTrue(req)
                req.wait()
                self.assertFalse(req)
                self.assertEqual(rmess, smess)
            else:
                req.Free()
            self.assertFalse(req)
            self.assertTrue (req == MPI.REQUEST_NULL)
            self.assertFalse(req != MPI.REQUEST_NULL)

    def testTestAll(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        requests = []
        for smess in messages:
            req = comm.issend(smess, rank)
            requests.append(req)
        flag, _ = self.RequestType.testall(requests)
        self.assertFalse(flag)
        flag, obj = self.RequestType.testall(requests)
        self.assertFalse(flag)
        self.assertEqual(obj, None)
        comm.barrier()
        for smess in messages:
            rmess = comm.recv(None, rank)
            self.assertEqual(rmess, smess)
        flag = False
        while not flag:
            flag, obj = self.RequestType.testall(requests)
        self.assertEqual(obj, [None]*len(messages))
        flag, _ = self.RequestType.testall(requests)
        self.assertTrue(flag)

    def testWaitAll(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        #
        requests = []
        for smess in messages:
            req = comm.issend(smess, dest)
            requests.append(req)
        flag, _ = self.RequestType.testall(requests)
        self.assertFalse(flag)
        flag, obj = self.RequestType.testall(requests)
        self.assertFalse(flag)
        self.assertEqual(obj, None)
        comm.barrier()
        for smess in messages:
            rmess = comm.recv(None, source)
            self.assertEqual(rmess, smess)
        obj = self.RequestType.waitall(requests)
        self.assertEqual(obj, [None]*len(messages))
        #
        requests1 = []
        for smess in messages:
            req = comm.issend(smess, dest)
            requests1.append(req)
        requests2 = []
        for _ in messages:
            req = comm.mprobe(source).irecv()
            requests2.append(req)
        statuses = [MPI.Status()]
        obj = self.RequestType.waitall(requests2, statuses)
        self.assertEqual(obj, messages)
        self.assertEqual(len(statuses), len(requests2))
        for status in statuses:
            self.assertEqual(status.source, source)
            self.assertEqual(status.tag, 0)
            self.assertTrue(status.Get_count() > 0)
        comm.barrier()
        statuses = (MPI.Status(),)
        self.RequestType.waitall(requests1, statuses)
        self.assertEqual(statuses[0].error, 0)

    def testSendrecv(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            rmess = self.COMM.sendrecv(smess, MPI.PROC_NULL, 0,
                                       None,  MPI.PROC_NULL, 0)
            self.assertEqual(rmess, None)
        if isinstance(self.COMM, pkl5.Comm):
            rbuf = MPI.Alloc_mem(32)
        else:
            rbuf = None
        for smess in messages:
            dest = (rank + 1) % size
            source = (rank - 1) % size
            rmess = self.COMM.sendrecv(None, dest,  0,
                                       None, source, 0)
            self.assertEqual(rmess, None)
            rmess = self.COMM.sendrecv(smess, dest,  0,
                                       None, source, 0)
            self.assertEqual(rmess, smess)
            status = MPI.Status()
            rmess = self.COMM.sendrecv(smess, dest,  42,
                                       rbuf, source, 42,
                                       status)
            self.assertEqual(status.source, source)
            self.assertEqual(status.tag, 42)
            self.assertEqual(status.error, 0)
        if rbuf is not None:
            MPI.Free_mem(rbuf)

    def testPingPong01(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for smess in messages:
            self.COMM.send(smess, MPI.PROC_NULL)
            rmess = self.COMM.recv(None, MPI.PROC_NULL, 0)
            self.assertEqual(rmess, None)
        if size == 1: return
        smess = None
        if rank == 0:
            self.COMM.send(smess, rank+1, 0)
            rmess = self.COMM.recv(None, rank+1, 0)
        elif rank == 1:
            rmess = self.COMM.recv(None, rank-1, 0)
            self.COMM.send(smess, rank-1, 0)
        else:
            rmess = smess
        self.assertEqual(rmess, smess)
        for smess in messages:
            if rank == 0:
                self.COMM.send(smess, rank+1, 0)
                rmess = self.COMM.recv(None, rank+1, 0)
            elif rank == 1:
                rmess = self.COMM.recv(None, rank-1, 0)
                self.COMM.send(smess, rank-1, 0)
            else:
                rmess = smess
            self.assertEqual(rmess, smess)

    def testIrecv(self):
        if isinstance(self.COMM, pkl5.Comm):
            self.assertRaises(
                RuntimeError,
                self.COMM.irecv,
                None,
                MPI.PROC_NULL,
                0,
            )

    def testProbe(self):
        comm = self.COMM.Dup()
        try:
            status = MPI.Status()
            flag = comm.iprobe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
            self.assertFalse(flag)
            for smess in messages:
                request = comm.issend(smess, comm.rank, 123)
                self.assertTrue(isinstance(request, self.RequestType))
                self.assertTrue(request != MPI.REQUEST_NULL)
                self.assertFalse(request == MPI.REQUEST_NULL)
                self.assertTrue(request == self.RequestType(request))
                self.assertFalse(request != self.RequestType(request))
                self.assertTrue(request != None)
                self.assertFalse(request == None)
                self.assertTrue(request)
                while not comm.iprobe(MPI.ANY_SOURCE, MPI.ANY_TAG, status): pass
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                comm.probe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                self.assertTrue(request)
                flag, obj = request.test()
                self.assertTrue(request)
                self.assertFalse(flag)
                self.assertEqual(obj, None)
                obj = comm.recv(None, comm.rank, 123)
                self.assertEqual(obj, smess)
                self.assertTrue(request)
                obj = request.wait()
                self.assertFalse(request)
                self.assertEqual(obj, None)
        finally:
            comm.Free()

    def testMProbe(self):
        comm = self.COMM.Dup()
        try:
            message = comm.mprobe(MPI.PROC_NULL)
            self.assertTrue(isinstance(message, self.MessageType))
            self.assertTrue(message == MPI.MESSAGE_NO_PROC)
            self.assertFalse(message != MPI.MESSAGE_NO_PROC)
            self.assertTrue(message != None)
            self.assertFalse(message == None)
            rmess = message.recv()
            self.assertTrue(message == MPI.MESSAGE_NULL)
            self.assertFalse(message != MPI.MESSAGE_NULL)
            self.assertTrue(rmess is None)

            message = comm.mprobe(MPI.PROC_NULL)
            self.assertTrue(isinstance(message, self.MessageType))
            self.assertTrue(message == MPI.MESSAGE_NO_PROC)
            self.assertFalse(message != MPI.MESSAGE_NO_PROC)
            request = message.irecv()
            self.assertTrue(message == MPI.MESSAGE_NULL)
            self.assertFalse(message != MPI.MESSAGE_NULL)
            self.assertTrue(request != MPI.REQUEST_NULL)
            self.assertFalse(request == MPI.REQUEST_NULL)
            rmess = request.wait()
            self.assertTrue(request == MPI.REQUEST_NULL)
            self.assertFalse(request != MPI.REQUEST_NULL)
            self.assertTrue(rmess is None)

            for smess in messages:
                request = comm.issend(smess, comm.rank, 123)
                message = comm.mprobe(comm.rank, 123)
                self.assertTrue(isinstance(message, self.MessageType))
                self.assertTrue(message == self.MessageType(message))
                self.assertFalse(message != self.MessageType(message))
                rmess = message.recv()
                self.assertEqual(rmess, smess)
                obj = request.wait()
                self.assertFalse(request)
                self.assertTrue(obj is None)
                flag, obj = request.test()
                self.assertTrue(flag)
                self.assertTrue(obj is None)
            for smess in messages:
                request = comm.issend(smess, comm.rank, 123)
                status = MPI.Status()
                message = comm.mprobe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                self.assertTrue(message)
                status = MPI.Status()
                rmess = message.recv(status)
                self.assertFalse(message)
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                self.assertEqual(rmess, smess)
                self.assertTrue(request)
                request.wait()
            for smess in messages:
                request = comm.issend(smess, comm.rank, 123)
                status = MPI.Status()
                message = comm.mprobe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                self.assertTrue(message)
                rreq = message.irecv()
                self.assertFalse(message)
                self.assertTrue(rreq)
                status = MPI.Status()
                rmess = rreq.wait(status)
                self.assertFalse(rreq)
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                self.assertEqual(rmess, smess)
                flag, obj = rreq.test()
                self.assertTrue(flag)
                self.assertTrue(obj is None)
                self.assertTrue(request)
                obj = request.wait()
                self.assertFalse(request)
                self.assertTrue(obj is None)
                flag, obj = request.test()
                self.assertTrue(flag)
                self.assertTrue(obj is None)
            for smess in messages:
                request = comm.issend(smess, comm.rank, 123)
                message = comm.mprobe(MPI.ANY_SOURCE, MPI.ANY_TAG)
                rreq = message.irecv()
                rreq.test()
                request.Free()
        finally:
            comm.Free()

    def testIMProbe(self):
        comm = self.COMM.Dup()
        try:
            status = MPI.Status()
            for smess in messages:
                message = comm.improbe(MPI.PROC_NULL)
                self.assertTrue(isinstance(message, self.MessageType))
                self.assertEqual(message, MPI.MESSAGE_NO_PROC)
            for smess in messages:
                message = comm.improbe(comm.rank, 123)
                self.assertEqual(message, None)
                request = comm.issend(smess, comm.rank, 123)
                while not comm.iprobe(comm.rank, 123): pass
                message = comm.improbe(comm.rank, 123)
                self.assertTrue(isinstance(message, self.MessageType))
                rmess = message.recv()
                self.assertEqual(rmess, smess)
                request.wait()
            for smess in messages:
                message = comm.improbe(comm.rank, 123)
                self.assertEqual(message, None)
                request = comm.issend(smess, comm.rank, 123)
                while not comm.iprobe(comm.rank, 123): pass
                message = comm.improbe(MPI.ANY_SOURCE, MPI.ANY_TAG, status)
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                self.assertTrue(message)
                rmess = message.recv()
                self.assertFalse(message)
                self.assertEqual(rmess, smess)
                self.assertTrue(request)
                request.wait()
                self.assertFalse(request)
        finally:
            comm.Free()

    def testMessageProbeIProbe(self):
        comm = self.COMM.Dup()
        try:
            status = MPI.Status()
            for smess in messages:
                request = comm.issend(smess, comm.rank, 123)
                message = self.MessageType.probe(comm, MPI.ANY_SOURCE, MPI.ANY_TAG, status)
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                self.assertTrue(message)
                rmess = message.recv()
                self.assertFalse(message)
                self.assertEqual(rmess, smess)
                self.assertTrue(request)
                request.wait()
                self.assertFalse(request)
            for smess in messages:
                message = self.MessageType.iprobe(comm, comm.rank, 123)
                self.assertEqual(message, None)
                request = comm.issend(smess, comm.rank, 123)
                while not comm.iprobe(comm.rank, 123): pass
                message = self.MessageType.iprobe(comm, MPI.ANY_SOURCE, MPI.ANY_TAG, status)
                self.assertEqual(status.source, comm.rank)
                self.assertEqual(status.tag, 123)
                self.assertTrue(message)
                rmess = message.recv()
                self.assertFalse(message)
                self.assertEqual(rmess, smess)
                self.assertTrue(request)
                request.wait()
                self.assertFalse(request)
        finally:
            comm.Free()

    def testSSendAndMProbe(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        if size == 1: return
        comm = self.COMM.Dup()
        try:
            for smess in messages:
                if rank == 0:
                    comm.ssend(smess, 1)
                    message = comm.mprobe(1)
                    rmess = message.recv()
                    self.assertEqual(rmess, smess)
                if rank == 1:
                    message = comm.mprobe(0)
                    rmess = message.recv()
                    comm.ssend(rmess, 0)
                    self.assertEqual(rmess, smess)
        finally:
            comm.Free()

    def testRequest(self):
        req = self.RequestType()
        self.assertFalse(req)
        self.assertEqual(req, self.RequestType())
        req = self.RequestType(MPI.REQUEST_NULL)
        self.assertFalse(req)
        self.assertEqual(req, MPI.REQUEST_NULL)
        self.assertEqual(req, self.RequestType())

    def testMessage(self):
        msg = self.MessageType()
        self.assertFalse(msg)
        self.assertEqual(msg, self.MessageType())
        msg = self.MessageType(MPI.MESSAGE_NULL)
        self.assertFalse(msg)
        self.assertEqual(msg, self.MessageType())
        msg = self.MessageType(MPI.MESSAGE_NO_PROC)
        self.assertTrue(msg)
        self.assertEqual(msg, MPI.MESSAGE_NO_PROC)
        self.assertEqual(msg, self.MessageType(MPI.MESSAGE_NO_PROC))
        self.assertNotEqual(msg, MPI.MESSAGE_NULL)

    def testBcastIntra(self, msglist=None, check=None):
        comm = self.COMM
        size = comm.Get_size()
        for smess in (msglist or messages):
            for root in range(size):
                rmess = comm.bcast(smess, root)
                if msglist and check:
                    self.assertTrue(check(rmess))
                else:
                    self.assertEqual(rmess, smess)

    def testBcastInter(self, msglist=None, check=None):
        basecomm = self.COMM
        size = basecomm.Get_size()
        rank = basecomm.Get_rank()
        if size == 1: return
        if rank < size // 2 :
            COLOR = 0
            local_leader = 0
            remote_leader = size // 2
        else:
            COLOR = 1
            local_leader = 0
            remote_leader = 0
        basecomm.Barrier()
        intracomm = basecomm.Split(COLOR, key=0)
        intercomm = MPI.Intracomm.Create_intercomm(
            intracomm,
            local_leader,
            basecomm,
            remote_leader
        )
        intracomm.Free()
        if isinstance(basecomm, pkl5.Intracomm):
            intercomm = pkl5.Intercomm(intercomm)
        rank = intercomm.Get_rank()
        size = intercomm.Get_size()
        rsize = intercomm.Get_remote_size()
        for smess in (msglist or messages)[:1]:
            intercomm.barrier()
            for color in [0, 1]:
                if COLOR == color:
                    for root in range(size):
                        if root == rank:
                            rmess = intercomm.bcast(smess, root=MPI.ROOT)
                        else:
                            rmess = intercomm.bcast(None, root=MPI.PROC_NULL)
                        self.assertEqual(rmess, None)
                else:
                    for root in range(rsize):
                        rmess = intercomm.bcast(None, root=root)
                        if msglist and check:
                            self.assertTrue(check(rmess))
                        else:
                            self.assertEqual(rmess, smess)
        if isinstance(intercomm, pkl5.Intercomm):
            bcast = intercomm.bcast
            rsize = intercomm.Get_remote_size()
            self.assertRaises(MPI.Exception, bcast, None, root=rsize)
        intercomm.Free()

    @unittest.skipIf(numpy is None, 'numpy')
    def testBigMPI(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        dest = (rank + 1) % size
        source = (rank - 1) % size
        bigmpi = self.bigmpi
        blocksizes = (
            63, 64, 65,
            (1<<12)-1,
            (1<<12),
            (1<<12)+1,
        )
        for blocksize in blocksizes:
            bigmpi.blocksize = blocksize
            a = numpy.empty(1024, dtype='i')
            b = numpy.empty(1024, dtype='i')
            c = numpy.empty(1024, dtype='i')
            a.fill(rank)
            b.fill(dest)
            c.fill(42)
            status = MPI.Status()
            smess = (a, b)
            rmess = comm.sendrecv(
                smess, dest,  42,
                None, source, 42,
                status,
            )
            self.assertTrue(numpy.all(rmess[0] == source))
            self.assertTrue(numpy.all(rmess[1] == rank))
            self.assertTrue(status.Get_elements(MPI.BYTE) > 0)
            comm.barrier()
            status = MPI.Status()
            smess = (a, b)
            request = comm.issend(smess, dest, 123)
            rmess = comm.mprobe(source, 123).irecv().wait(status)
            self.assertTrue(numpy.all(rmess[0] == source))
            self.assertTrue(numpy.all(rmess[1] == rank))
            self.assertTrue(status.Get_elements(MPI.BYTE) > 0)
            request.Free()
            comm.barrier()
            check = lambda x: numpy.all(x == 42)
            self.testBcastIntra([c, c], check)
            self.testBcastInter([c, c], check)
            check2 = lambda x: check(x[0]) and check(x[1])
            self.testBcastIntra([(c, c.copy())], check2)
            self.testBcastInter([(c, c.copy())], check2)


class BaseTestPKL5(object):
    CommType = pkl5.Intracomm
    MessageType = pkl5.Message
    RequestType = pkl5.Request

    def setUp(self):
        super(BaseTestPKL5, self).setUp()
        self.pickle_prev = pkl5.pickle
        self.pickle = pkl5.Pickle()
        self.pickle.THRESHOLD = 0
        pkl5.pickle = self.pickle

    def tearDown(self):
        super(BaseTestPKL5, self).tearDown()
        pkl5.pickle = self.pickle_prev

    @unittest.skipIf(numpy is None, 'numpy')
    def testPickle5(self):
        comm = self.COMM
        rank = comm.Get_rank()
        pickle = self.pickle
        protocols = list(range(-2, pickle.PROTOCOL+1))
        for protocol in [None] + protocols:
            pickle.PROTOCOL = protocol
            for threshold in (-1, 0, 64, 256):
                pickle.THRESHOLD = threshold
                for slen in (0, 32, 64, 128, 256, 512):
                    sobj = numpy.empty(slen, dtype='i')
                    sobj.fill(rank)
                    #
                    robj = comm.sendrecv(
                        sobj, rank, 42,
                        None, rank, 42)
                    self.assertTrue(numpy.all(sobj==robj))
                    #
                    data, bufs = pickle.dumps(sobj)
                    robj = pickle.loads(data, bufs)
                    self.assertTrue(numpy.all(sobj==robj))
                    if protocol is None:
                        protocol = MPI.Pickle().PROTOCOL
                    if protocol < 0:
                        protocol = pkl5._PROTOCOL
                    if protocol >= 5 and sobj.nbytes >= threshold:
                        self.assertEqual(len(bufs), 1)
                    else:
                        self.assertEqual(len(bufs), 0)


class TestMPISelf(BaseTest, unittest.TestCase):
    COMM = MPI.COMM_SELF

class TestMPIWorld(BaseTest, unittest.TestCase):
    COMM = MPI.COMM_WORLD

class TestPKL5Self(BaseTestPKL5, TestMPISelf):
    pass

class TestPKL5World(BaseTestPKL5, TestMPIWorld):
    pass


if __name__ == '__main__':
    unittest.main()



import sys, os
import optparse
import unittest

def getoptionparser():
    parser = optparse.OptionParser()

    parser.add_option("-q", "--quiet",
                      action="store_const", const=0, dest="verbose", default=1,
                      help="minimal output")
    parser.add_option("-v", "--verbose",
                      action="store_const", const=2, dest="verbose", default=1,
                      help="verbose output")
    parser.add_option("-i", "--include", type="string",
                      action="append",  dest="include", default=[],
                      help="include tests matching PATTERN", metavar="PATTERN")
    parser.add_option("-e", "--exclude", type="string",
                      action="append", dest="exclude", default=[],
                      help="exclude tests matching PATTERN", metavar="PATTERN")
    parser.add_option("-f", "--failfast",
                      action="store_true", dest="failfast", default=False,
                      help="stop on first failure")
    parser.add_option("-c", "--catch",
                      action="store_true", dest="catchbreak", default=False,
                      help="catch Control-C and display results")
    parser.add_option("-k", "--pattern", type="string",
                      action="append", dest="patterns", default=[],
                      help="only run tests which match the given substring")
    parser.add_option("--no-builddir",
                      action="store_false", dest="builddir", default=True,
                      help="disable testing from build directory")
    parser.add_option("--path", type="string",
                      action="append", dest="path", default=[],
                      help="prepend PATH to sys.path", metavar="PATH")
    parser.add_option("--refleaks", type="int",
                      action="store", dest="repeats", default=3,
                      help="run tests REPEAT times in a loop to catch leaks",
                      metavar="REPEAT")
    parser.add_option("--threads",
                      action="store_true", dest="threads", default=None,
                      help="initialize MPI with thread support")
    parser.add_option("--no-threads",
                      action="store_false", dest="threads", default=None,
                      help="initialize MPI without thread support")
    parser.add_option("--thread-level", type="choice",
                      choices=["single", "funneled", "serialized", "multiple"],
                      action="store", dest="thread_level", default=None,
                      help="initialize MPI with required thread support")
    parser.add_option("--mpe",
                      action="store_true", dest="mpe", default=False,
                      help="use MPE for MPI profiling")
    parser.add_option("--vt",
                      action="store_true", dest="vt", default=False,
                      help="use VampirTrace for MPI profiling")
    parser.add_option("--cupy",
                      action="store_true", dest="cupy", default=False,
                      help="enable testing with CuPy arrays")
    parser.add_option("--no-cupy",
                      action="store_false", dest="cupy", default=False,
                      help="disable testing with CuPy arrays")
    parser.add_option("--numba",
                      action="store_true", dest="numba", default=False,
                      help="enable testing with Numba arrays")
    parser.add_option("--no-numba",
                      action="store_false", dest="numba", default=False,
                      help="disable testing with Numba arrays")
    parser.add_option("--no-numpy",
                      action="store_false", dest="numpy", default=True,
                      help="disable testing with NumPy arrays")
    parser.add_option("--no-array",
                      action="store_false", dest="array", default=True,
                      help="disable testing with builtin array module")
    parser.add_option("--no-skip-mpi",
                      action="store_false", dest="skip_mpi", default=True,
                      help="disable known failures with backend MPI")
    return parser

def getbuilddir():
    from distutils.util import get_platform
    plat_name, (x, y) = get_platform(), sys.version_info[:2]
    s = os.path.join("build", "lib.%s-%d.%d" % (plat_name, x, y))
    if hasattr(sys, 'gettotalrefcount'): s += '-pydebug'
    return s

def setup_python(options):
    rootdir = os.path.dirname(os.path.dirname(__file__))
    builddir = os.path.join(rootdir, getbuilddir())
    if options.builddir and os.path.exists(builddir):
        sys.path.insert(0, builddir)
    if options.path:
        path = options.path[:]
        path.reverse()
        for p in path:
            sys.path.insert(0, p)

def setup_unittest(options):
    from unittest import TestSuite
    try:
        from unittest.runner import _WritelnDecorator
    except ImportError:
        from unittest import _WritelnDecorator
    #
    writeln_orig = _WritelnDecorator.writeln
    def writeln(self, message=''):
        try: self.stream.flush()
        except: pass
        writeln_orig(self, message)
        try: self.stream.flush()
        except: pass
    _WritelnDecorator.writeln = writeln

def import_package(options, pkgname):
    #
    if not options.cupy:
        sys.modules['cupy'] = None
    if not options.numba:
        sys.modules['numba'] = None
    if not options.numpy:
        sys.modules['numpy'] = None
    if not options.array:
        sys.modules['array'] = None
    #
    package = __import__(pkgname)
    #
    import mpi4py.rc
    if options.threads is not None:
        mpi4py.rc.threads = options.threads
    if options.thread_level is not None:
        mpi4py.rc.thread_level = options.thread_level
    if options.mpe:
        mpi4py.profile('mpe', logfile='runtests-mpi4py')
    if options.vt:
        mpi4py.profile('vt', logfile='runtests-mpi4py')
    import mpi4py.MPI
    #
    return package

def getprocessorinfo():
    from mpi4py import MPI
    rank = MPI.COMM_WORLD.Get_rank()
    name = MPI.Get_processor_name()
    return (rank, name)

def getlibraryinfo():
    from mpi4py import MPI
    info = "MPI %d.%d" % MPI.Get_version()
    name, version = MPI.get_vendor()
    if name != "unknown":
        info += (" (%s %s)" % (name, '%d.%d.%d' % version))
    return info

def getpythoninfo():
    x, y = sys.version_info[:2]
    return ("Python %d.%d (%s)" % (x, y, sys.executable))

def getpackageinfo(pkg):
    return ("%s %s (%s)" % (pkg.__name__,
                            pkg.__version__,
                            pkg.__path__[0]))

def writeln(message='', endl='\n'):
    sys.stderr.flush()
    sys.stderr.write(message+endl)
    sys.stderr.flush()

def print_banner(options, package):
    r, n = getprocessorinfo()
    fmt = "[%d@%s] %s"
    if options.verbose:
        writeln(fmt % (r, n, getpythoninfo()))
        writeln(fmt % (r, n, getlibraryinfo()))
        writeln(fmt % (r, n, getpackageinfo(package)))

def load_tests(options, args):
    # Find tests
    import re, glob
    testsuitedir = os.path.dirname(__file__)
    sys.path.insert(0, testsuitedir)
    pattern = 'test_*.py'
    wildcard = os.path.join(testsuitedir, pattern)
    testfiles = glob.glob(wildcard)
    include = exclude = None
    if options.include:
        include = re.compile('|'.join(options.include)).search
    if options.exclude:
        exclude = re.compile('|'.join(options.exclude)).search
    testnames = []
    for testfile in testfiles:
        filename = os.path.basename(testfile)
        testname = os.path.splitext(filename)[0]
        if ((exclude and exclude(testname)) or
            (include and not include(testname))):
            continue
        testnames.append(testname)
    testnames.sort()
    # Handle options
    if not options.cupy:
        sys.modules['cupy'] = None
    if not options.numba:
        sys.modules['numba'] = None
    if not options.numpy:
        sys.modules['numpy'] = None
    if not options.array:
        sys.modules['array'] = None
    if not options.skip_mpi:
        import mpiunittest
        mpiunittest.skipMPI = lambda p, *c: lambda f: f
    # Load tests and populate suite
    testloader = unittest.TestLoader()
    if options.patterns:
        testloader.testNamePatterns = [
            ('*%s*' % p) if ('*' not in p) else p
            for p in options.patterns]
    testsuite = unittest.TestSuite()
    for testname in testnames:
        module = __import__(testname)
        for arg in args:
            try:
                cases = testloader.loadTestsFromNames((arg,), module)
            except AttributeError:
                continue
            testsuite.addTests(cases)
        if not args:
            cases = testloader.loadTestsFromModule(module)
            testsuite.addTests(cases)
    return testsuite

def run_tests(options, testsuite, runner=None):
    if runner is None:
        runner = unittest.TextTestRunner()
        runner.verbosity = options.verbose
        runner.failfast = options.failfast
    if options.catchbreak:
        unittest.installHandler()
    result = runner.run(testsuite)
    return result.wasSuccessful()

def test_refleaks(options, args):
    from sys import gettotalrefcount
    from gc import collect
    testsuite = load_tests(options, args)
    testsuite._cleanup =  False
    for case in testsuite:
        case._cleanup = False
    class EmptyIO(object):
        def write(self, *args):
            pass
    runner = unittest.TextTestRunner(stream=EmptyIO(), verbosity=0)
    rank, name = getprocessorinfo()
    r1 = r2 = 0
    repeats = options.repeats
    while repeats:
        collect()
        r1 = gettotalrefcount()
        run_tests(options, testsuite, runner)
        collect()
        r2 = gettotalrefcount()
        leaks = r2-r1
        if leaks and repeats < options.repeats:
            writeln('[%d@%s] refleaks:  (%d - %d) --> %d'
                    % (rank, name, r2, r1, leaks))
        repeats -= 1

def abort(code=1):
    from mpi4py import MPI
    MPI.COMM_WORLD.Abort(code)

def shutdown(success):
    from mpi4py import MPI

def main(args=None):
    pkgname = 'mpi4py'
    parser = getoptionparser()
    (options, args) = parser.parse_args(args)
    setup_python(options)
    setup_unittest(options)
    package = import_package(options, pkgname)
    print_banner(options, package)
    testsuite = load_tests(options, args)
    success = run_tests(options, testsuite)
    if not success and options.failfast: abort()
    if success and hasattr(sys, 'gettotalrefcount'):
        test_refleaks(options, args)
    shutdown(success)
    return not success

if __name__ == '__main__':
    import sys
    sys.dont_write_bytecode = True
    sys.exit(main())



import sys
from mpi4py import MPI
try:
    from collections import OrderedDict
except ImportError:
    OrderedDict = dict
try:
    import array
except ImportError:
    array = None
try:
    import numpy
except ImportError:
    numpy = None
try:
    import cupy
except ImportError:
    cupy = None
try:
    import numba
    import numba.cuda
    from distutils.version import StrictVersion
    numba_version = StrictVersion(numba.__version__).version
    if numba_version < (0, 48):
        import warnings
        warnings.warn('To test Numba GPU arrays, use Numba v0.48.0+.',
                      RuntimeWarning)
        numba = None
except ImportError:
    numba = None


__all__ = ['allclose', 'subTest']

def allclose(a, b, rtol=1.e-5, atol=1.e-8):
    try: iter(a)
    except TypeError: a = [a]
    try: iter(b)
    except TypeError: b = [b]
    for x, y in zip(a, b):
        if abs(x-y) > (atol + rtol * abs(y)):
            return False
    return True

def make_typemap(entries):
    typemap = OrderedDict(entries)
    for typecode, datatype in entries:
        if datatype == MPI.DATATYPE_NULL:
            del typemap[typecode]
    return typemap

TypeMap = make_typemap([
    ('b', MPI.SIGNED_CHAR),
    ('h', MPI.SHORT),
    ('i', MPI.INT),
    ('l', MPI.LONG),
    ('q', MPI.LONG_LONG),
    ('f', MPI.FLOAT),
    ('d', MPI.DOUBLE),
    ('g', MPI.LONG_DOUBLE),
])

TypeMapBool = make_typemap([
    ('?', MPI.C_BOOL),
])

TypeMapInteger = make_typemap([
    ('b', MPI.SIGNED_CHAR),
    ('h', MPI.SHORT),
    ('i', MPI.INT),
    ('l', MPI.LONG),
    ('q', MPI.LONG_LONG),
])

TypeMapUnsigned = make_typemap([
    ('B', MPI.UNSIGNED_CHAR),
    ('H', MPI.UNSIGNED_SHORT),
    ('I', MPI.UNSIGNED_INT),
    ('L', MPI.UNSIGNED_LONG),
    ('Q', MPI.UNSIGNED_LONG_LONG),
])

TypeMapFloat = make_typemap([
    ('f', MPI.FLOAT),
    ('d', MPI.DOUBLE),
    ('g', MPI.LONG_DOUBLE),
])

TypeMapComplex = make_typemap([
    ('F', MPI.C_FLOAT_COMPLEX),
    ('D', MPI.C_DOUBLE_COMPLEX),
    ('G', MPI.C_LONG_DOUBLE_COMPLEX),
])


ArrayBackends = []

def add_backend(cls):
    ArrayBackends.append(cls)
    return cls

class BaseArray(object):

    backend = None

    TypeMap = TypeMap.copy()
    TypeMap.pop('g', None)
    if sys.version_info[:2] < (3, 3):
        TypeMap.pop('q', None)

    def __len__(self):
        return len(self.array)

    def __getitem__(self, i):
        return self.array[i]

    def __setitem__(self, i, v):
        self.array[i] = v

    @property
    def mpidtype(self):
        try:
            return self.TypeMap[self.typecode]
        except KeyError:
            return MPI.DATATYPE_NULL

    def as_raw(self):
        return self.array

    def as_mpi(self):
        return (self.as_raw(), self.mpidtype)

    def as_mpi_c(self, count):
        return (self.as_raw(), count, self.mpidtype)

    def as_mpi_v(self, cnt, dsp):
        return (self.as_raw(), (cnt, dsp), self.mpidtype)


if array is not None:

    def product(seq):
        res = 1
        for s in seq:
            res = res * s
        return res

    def mkshape(shape):
        return tuple([int(s) for s in shape])

    @add_backend
    class ArrayArray(BaseArray):

        backend = 'array'

        def __init__(self, arg, typecode, shape=None):
            if isinstance(arg, (int, float)):
                if shape is None:
                    shape = ()
                else:
                    try:
                        shape = mkshape(shape)
                    except TypeError:
                        shape = (int(shape),)
                size = product(shape)
                arg = [arg] * size
            else:
                size = len(arg)
                if shape is None:
                    shape = (size,)
                else:
                    shape = mkshape(shape)
                assert size == product(shape)
            self.array = array.array(typecode, arg)

        @property
        def address(self):
            return self.array.buffer_info()[0]

        @property
        def typecode(self):
            return self.array.typecode

        @property
        def itemsize(self):
            return self.array.itemsize

        @property
        def flat(self):
            return self.array

        @property
        def size(self):
            return self.array.buffer_info()[1]


if numpy is not None:

    @add_backend
    class ArrayNumPy(BaseArray):

        backend = 'numpy'

        TypeMap = make_typemap([])
        #TypeMap.update(TypeMapBool)
        TypeMap.update(TypeMapInteger)
        #TypeMap.update(TypeMapUnsigned)
        TypeMap.update(TypeMapFloat)
        TypeMap.update(TypeMapComplex)

        def __init__(self, arg, typecode, shape=None):
            if isinstance(arg, (int, float, complex)):
                if shape is None: shape = ()
            else:
                if shape is None: shape = len(arg)
            self.array = numpy.zeros(shape, typecode)
            if isinstance(arg, (int, float, complex)):
                self.array.fill(arg)
            else:
                self.array[:] = numpy.asarray(arg, typecode)

        @property
        def address(self):
            return self.array.__array_interface__['data'][0]

        @property
        def typecode(self):
            return self.array.dtype.char

        @property
        def itemsize(self):
            return self.array.itemsize

        @property
        def flat(self):
            return self.array.flat

        @property
        def size(self):
            return self.array.size


try:
    import dlpackimpl as dlpack
except ImportError:
    dlpack = None

class BaseDLPackCPU(object):

    def __dlpack_device__(self):
        return (dlpack.DLDeviceType.kDLCPU, 0)

    def __dlpack__(self, stream=None):
        assert stream is None
        capsule = dlpack.make_py_capsule(self.array)
        return capsule

    def as_raw(self):
        return self


if dlpack is not None and array is not None:

    @add_backend
    class DLPackArray(BaseDLPackCPU, ArrayArray):

        backend = 'dlpack-array'

        def __init__(self, arg, typecode, shape=None):
            super(DLPackArray, self).__init__(arg, typecode, shape)


if dlpack is not None and numpy is not None:

    @add_backend
    class DLPackNumPy(BaseDLPackCPU, ArrayNumPy):

        backend = 'dlpack-numpy'

        def __init__(self, arg, typecode, shape=None):
            super(DLPackNumPy, self).__init__(arg, typecode, shape)


def typestr(typecode, itemsize):
    typestr = ''
    if sys.byteorder == 'little':
        typestr += '<'
    if sys.byteorder == 'big':
        typestr += '>'
    if typecode in '?':
        typestr += 'b'
    if typecode in 'bhilq':
        typestr += 'i'
    if typecode in 'BHILQ':
        typestr += 'u'
    if typecode in 'fdg':
        typestr += 'f'
    if typecode in 'FDG':
        typestr += 'c'
    typestr += str(itemsize)
    return typestr


class BaseFakeGPUArray(object):

    def set_interface(self, shape, readonly=False):
        self.__cuda_array_interface__ = dict(
            version = 0,
            data    = (self.address, readonly),
            typestr = typestr(self.typecode, self.itemsize),
            shape   = shape,
        )

    def as_raw(self):
        return self


if array is not None:

    @add_backend
    class FakeGPUArrayBasic(BaseFakeGPUArray, ArrayArray):

        def __init__(self, arg, typecode, shape=None, readonly=False):
            super(FakeGPUArrayBasic, self).__init__(arg, typecode, shape)
            self.set_interface((len(self),), readonly)


if numpy is not None:

    @add_backend
    class FakeGPUArrayNumPy(BaseFakeGPUArray, ArrayNumPy):

        def __init__(self, arg, typecode, shape=None, readonly=False):
            super(FakeGPUArrayNumPy, self).__init__(arg, typecode, shape)
            self.set_interface(self.array.shape, readonly)


if cupy is not None:

    @add_backend
    class GPUArrayCuPy(BaseArray):

        backend = 'cupy'

        TypeMap = make_typemap([])
        #TypeMap.update(TypeMapBool)
        TypeMap.update(TypeMapInteger)
        #TypeMap.update(TypeMapUnsigned)
        TypeMap.update(TypeMapFloat)
        TypeMap.update(TypeMapComplex)
        try:
            cupy.array(0, 'g')
        except ValueError:
            TypeMap.pop('g', None)
        try:
            cupy.array(0, 'G')
        except ValueError:
            TypeMap.pop('G', None)

        def __init__(self, arg, typecode, shape=None, readonly=False):
            if isinstance(arg, (int, float, complex)):
                if shape is None: shape = ()
            else:
                if shape is None: shape = len(arg)
            self.array = cupy.zeros(shape, typecode)
            if isinstance(arg, (int, float, complex)):
                self.array.fill(arg)
            else:
                self.array[:] = cupy.asarray(arg, typecode)

        @property
        def address(self):
            return self.array.__cuda_array_interface__['data'][0]

        @property
        def typecode(self):
            return self.array.dtype.char

        @property
        def itemsize(self):
            return self.array.itemsize

        @property
        def flat(self):
            return self.array.ravel()

        @property
        def size(self):
            return self.array.size

        def as_raw(self):
            cupy.cuda.get_current_stream().synchronize()
            return self.array


if cupy is not None:

    # Note: we do not create a BaseDLPackGPU class because each GPU library
    # has its own way to get device ID etc, so we have to reimplement the
    # DLPack support anyway

    @add_backend
    class DLPackCuPy(GPUArrayCuPy):

        backend = 'dlpack-cupy'
        has_dlpack = None
        dev_type = None

        def __init__(self, arg, typecode, shape=None):
            super().__init__(arg, typecode, shape)
            self.has_dlpack = hasattr(self.array, '__dlpack_device__')
            # TODO(leofang): test CUDA managed memory?
            if cupy.cuda.runtime.is_hip:
                self.dev_type = dlpack.DLDeviceType.kDLROCM
            else:
                self.dev_type = dlpack.DLDeviceType.kDLCUDA

        def __dlpack_device__(self):
            if self.has_dlpack:
                return self.array.__dlpack_device__()
            else:
                return (self.dev_type, self.array.device.id)

        def __dlpack__(self, stream=None):
            cupy.cuda.get_current_stream().synchronize()
            if self.has_dlpack:
                return self.array.__dlpack__(stream=-1)
            else:
                return self.array.toDlpack()

        def as_raw(self):
            return self


if numba is not None:

    @add_backend
    class GPUArrayNumba(BaseArray):

        backend = 'numba'

        TypeMap = make_typemap([])
        #TypeMap.update(TypeMapBool)
        TypeMap.update(TypeMapInteger)
        #TypeMap.update(TypeMapUnsigned)
        TypeMap.update(TypeMapFloat)
        TypeMap.update(TypeMapComplex)

        # one can allocate arrays with those types,
        # but the Numba compiler doesn't support them...
        TypeMap.pop('g', None)
        TypeMap.pop('G', None)

        def __init__(self, arg, typecode, shape=None, readonly=False):
            if isinstance(arg, (int, float, complex)):
                if shape is None: shape = ()
            else:
                if shape is None: shape = len(arg)
            self.array = numba.cuda.device_array(shape, typecode)
            if isinstance(arg, (int, float, complex)):
                if self.array.size > 0:
                    self.array[:] = arg
            elif arg == [] or arg == ():
                self.array = numba.cuda.device_array(0, typecode)
            else:
                if self.array.size > 0:
                    self.array[:] = numba.cuda.to_device(arg)

#        def __getitem__(self, i):
#            if isinstance(i, slice):
#                return self.array[i]
#            elif i < self.array.size:
#                return self.array[i]
#            else:
#                raise StopIteration

        @property
        def address(self):
            return self.array.__cuda_array_interface__['data'][0]

        @property
        def typecode(self):
            return self.array.dtype.char

        @property
        def itemsize(self):
            return self.array.dtype.itemsize

        @property
        def flat(self):
            if self.array.ndim <= 1:
                return self.array
            else:
                return self.array.ravel()

        @property
        def size(self):
            return self.array.size

        def as_raw(self):
            # numba by default always runs on the legacy default stream
            numba.cuda.default_stream().synchronize()
            return self.array


def subTest(case, skip=(), skiptypecode=()):
    for array in ArrayBackends:
        if array.backend == skip: continue
        if array.backend in skip: continue
        for typecode in array.TypeMap:
            if typecode == skiptypecode: continue
            if typecode in skiptypecode: continue
            with case.subTest(backend=array.backend, typecode=typecode):
                try:
                    yield array, typecode
                except GeneratorExit:
                    return



from mpi4py import MPI
import ctypes
import os

_libdir = os.path.dirname(__file__)

if MPI._sizeof(MPI.Comm) == ctypes.sizeof(ctypes.c_int):
    MPI_Comm = ctypes.c_int
else:
    MPI_Comm = ctypes.c_void_p
_lib = ctypes.CDLL(os.path.join(_libdir, "libhelloworld.so"))
_lib.sayhello.restype = None
_lib.sayhello.argtypes = [MPI_Comm]

def sayhello(comm):
    comm_ptr = MPI._addressof(comm)
    comm_val = MPI_Comm.from_address(comm_ptr)
    _lib.sayhello(comm_val)



from mpi4py import MPI
try:
    import numpy
except ImportError:
    raise SystemExit

# extract the section a[0:6:2, 0:5:2] and store it in e[:,:]

a = numpy.empty((6, 5), dtype=float, order='fortran')
e = numpy.empty((3, 3), dtype=float, order='fortran')
a.flat = numpy.arange(a.size, dtype=float)

lb, sizeofdouble = MPI.DOUBLE.Get_extent()

# create datatype for a 1D section
oneslice = MPI.DOUBLE.Create_vector(3, 1, 2)

# create datatype for a 2D section
twoslice = oneslice.Create_hvector(3, 1, 12*sizeofdouble)
twoslice.Commit()

# send and recv on same process
myrank = MPI.COMM_WORLD.Get_rank()
status = MPI.Status()
MPI.COMM_WORLD.Sendrecv([a, 1, twoslice], myrank, 0,
                        (e, MPI.DOUBLE), myrank, 0, status)

assert numpy.allclose(a[::2, ::2], e)
assert status.Get_count(twoslice) == 1
assert status.Get_count(MPI.DOUBLE) == e.size

oneslice.Free()
twoslice.Free()




from mpi4py import MPI
import mpiunittest as unittest
import sys, types

ModuleType = type(MPI)
ClassType = type(MPI.Comm)
FunctionType = type(MPI.Init)
MethodDescrType = type(MPI.Comm.Get_rank)
GetSetDescrType = type(MPI.Comm.rank)

def getdocstr(mc, docstrings, namespace=None):
    name = getattr(mc, '__name__', None)
    if name is None: return
    if name in ('__builtin__', 'builtins'): return
    if name.startswith('_'): return
    if namespace: name = '%s.%s' % (namespace, name)
    if type(mc) in (ModuleType, ClassType):
        doc = getattr(mc, '__doc__', None)
        if doc == "<undocumented>": return
        docstrings[name] = doc
        for k, v in vars(mc).items():
            if isinstance(v, staticmethod):
                v = v.__get__(object)
            getdocstr(v, docstrings, name)
    elif type(mc) in (FunctionType, MethodDescrType, GetSetDescrType):
        doc = getattr(mc, '__doc__', None)
        if doc == "<undocumented>": return
        docstrings[name] = doc

class TestDoc(unittest.TestCase):

    @unittest.skipIf(hasattr(sys, 'pypy_version_info'), 'pypy')
    def testDoc(self):
        missing = False
        docs = { }
        getdocstr(MPI, docs)
        for k in docs:
            if not k.startswith('_'):
                doc = docs[k]
                if doc is None:
                    print ("'%s': missing docstring" % k)
                    missing = True
                else:
                    doc = doc.strip()
                    if not doc:
                        print ("'%s': empty docstring" % k)
                        missing = True
                    if 'mpi4py.MPI' in doc:
                        print ("'%s': bad format docstring" % k)
        self.assertFalse(missing)


if __name__ == '__main__':
    unittest.main()



from mpi4py import MPI
import mpiunittest as unittest
try:
    import socket
except ImportError:
    socket = None


def ch4_ucx():
    return 'ch4:ucx' in MPI.Get_library_version()

def ch4_ofi():
    return 'ch4:ofi' in MPI.Get_library_version()

def appnum():
    if MPI.APPNUM == MPI.KEYVAL_INVALID: return None
    return MPI.COMM_WORLD.Get_attr(MPI.APPNUM)

def badport():
    if MPI.get_vendor()[0] != 'MPICH':
        return False
    try:
        port = MPI.Open_port()
        MPI.Close_port(port)
    except:
        port = ""
    return port == ""

@unittest.skipMPI('mpich', badport())
@unittest.skipMPI('openmpi(<2.0.0)')
@unittest.skipMPI('MVAPICH2')
@unittest.skipMPI('msmpi(<8.1.0)')
class TestDPM(unittest.TestCase):

    message = [
        None,
        True, False,
        -7, 0, 7,
        -2**63+1, 2**63-1,
        -2.17, 0.0, 3.14,
        1+2j, 2-3j,
        'mpi4py',
        (1, 2, 3),
        [1, 2, 3],
        {1:2},
    ]

    @unittest.skipMPI('mpich', appnum() is None)
    @unittest.skipMPI('MPICH2', appnum() is None)
    @unittest.skipMPI('MPICH1', appnum() is None)
    @unittest.skipMPI('msmpi(<8.1.0)', appnum() is None)
    @unittest.skipMPI('PlatformMPI')
    def testNamePublishing(self):
        rank = MPI.COMM_WORLD.Get_rank()
        service = "mpi4py-%d" % rank
        port = MPI.Open_port()
        MPI.Publish_name(service, port)
        found =  MPI.Lookup_name(service)
        self.assertEqual(port, found)
        MPI.Unpublish_name(service, port)
        MPI.Close_port(port)

    @unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, 'mpi-world-size<2')
    @unittest.skipMPI('mpich(==3.4.1)', ch4_ofi())
    def testAcceptConnect(self):
        comm_self  = MPI.COMM_SELF
        comm_world = MPI.COMM_WORLD
        wsize = comm_world.Get_size()
        wrank = comm_world.Get_rank()
        group_world = comm_world.Get_group()
        group = group_world.Excl([0])
        group_world.Free()
        comm = comm_world.Create(group)
        group.Free()
        if wrank == 0:
            self.assertEqual(comm, MPI.COMM_NULL)
        else:
            self.assertNotEqual(comm, MPI.COMM_NULL)
            self.assertEqual(comm.size, comm_world.size-1)
            self.assertEqual(comm.rank, comm_world.rank-1)
        if wrank == 0:
            port = MPI.Open_port()
            comm_world.send(port, dest=1)
            intercomm = comm_self.Accept(port)
            self.assertEqual(intercomm.remote_size, comm_world.size-1)
            self.assertEqual(intercomm.size, 1)
            self.assertEqual(intercomm.rank, 0)
            MPI.Close_port(port)
        else:
            if wrank == 1:
                port = comm_world.recv(source=0)
            else:
                port = None
            intercomm = comm.Connect(port, root=0)
            self.assertEqual(intercomm.remote_size, 1)
            self.assertEqual(intercomm.size, comm_world.size-1)
            self.assertEqual(intercomm.rank, comm.rank)
            comm.Free()
        if wrank == 0:
            message = TestDPM.message
            root = MPI.ROOT
        else:
            message = None
            root = 0
        message = intercomm.bcast(message, root)
        if wrank == 0:
            self.assertEqual(message, None)
        else:
            self.assertEqual(message, TestDPM.message)
        intercomm.Free()

    @unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, 'mpi-world-size<2')
    def testConnectAccept(self):
        comm_self  = MPI.COMM_SELF
        comm_world = MPI.COMM_WORLD
        wsize = comm_world.Get_size()
        wrank = comm_world.Get_rank()
        group_world = comm_world.Get_group()
        group = group_world.Excl([0])
        group_world.Free()
        comm = comm_world.Create(group)
        group.Free()
        if wrank == 0:
            self.assertEqual(comm, MPI.COMM_NULL)
        else:
            self.assertNotEqual(comm, MPI.COMM_NULL)
            self.assertEqual(comm.size, comm_world.size-1)
            self.assertEqual(comm.rank, comm_world.rank-1)
        if wrank == 0:
            port = comm_world.recv(source=1)
            intercomm = comm_self.Connect(port)
            self.assertEqual(intercomm.remote_size, comm_world.size-1)
            self.assertEqual(intercomm.size, 1)
            self.assertEqual(intercomm.rank, 0)
        else:
            if wrank == 1:
                port = MPI.Open_port()
                comm_world.send(port, dest=0)
            else:
                port = None
            intercomm = comm.Accept(port, root=0)
            if wrank == 1:
                MPI.Close_port(port)
            self.assertEqual(intercomm.remote_size, 1)
            self.assertEqual(intercomm.size, comm_world.size-1)
            self.assertEqual(intercomm.rank, comm.rank)
            comm.Free()
        if wrank == 0:
            message = TestDPM.message
            root = MPI.ROOT
        else:
            message = None
            root = 0
        message = intercomm.bcast(message, root)
        if wrank == 0:
            self.assertEqual(message, None)
        else:
            self.assertEqual(message, TestDPM.message)
        intercomm.Free()

    @unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, 'mpi-world-size<2')
    @unittest.skipIf(socket is None, 'socket')
    def testJoin(self):
        size = MPI.COMM_WORLD.Get_size()
        rank = MPI.COMM_WORLD.Get_rank()
        server = client = address = None
        host = socket.gethostname()
        addresses = socket.getaddrinfo(host, None, 0, socket.SOCK_STREAM)
        address_families = [ a[0] for a in addresses ]
        # if both INET and INET6 are available, don't assume the order
        # is the same on both server and client. Select INET if available.
        if socket.AF_INET in address_families:
            socket_family = socket.AF_INET
        elif socket.AF_INET6 in address_families:
            socket_family = socket.AF_INET6
        elif address_families:
            # allow for AF_UNIX (or other families)
            socket_family = address_families[0]
        else:
            self.skipTest("socket")
        # create server/client sockets
        if rank == 0: # server
            server = socket.socket(socket_family, socket.SOCK_STREAM)
            server.bind((host, 0))
            server.listen(0)
        if rank == 1: # client
            client = socket.socket(socket_family, socket.SOCK_STREAM)
        # communicate address
        if rank == 0:
            address = server.getsockname()
            MPI.COMM_WORLD.ssend(address, 1)
        if rank == 1:
            address = MPI.COMM_WORLD.recv(None, 0)
        MPI.COMM_WORLD.Barrier()
        # stablish client/server connection
        connected = False
        if rank == 0: # server
            client = server.accept()[0]
            server.close()
        if rank == 1: # client
            try:
                client.connect(address)
                connected = True
            except socket.error:
                raise
        connected = MPI.COMM_WORLD.bcast(connected, root=1)
        # test Comm.Join()
        MPI.COMM_WORLD.Barrier()
        if client:
            fd = client.fileno()
            intercomm = MPI.Comm.Join(fd)
            client.close()
            if intercomm != MPI.COMM_NULL:
                self.assertEqual(intercomm.remote_size, 1)
                self.assertEqual(intercomm.size, 1)
                self.assertEqual(intercomm.rank, 0)
                if rank == 0:
                    message = TestDPM.message
                    root = MPI.ROOT
                else:
                    message = None
                    root = 0
                message = intercomm.bcast(message, root)
                if rank == 0:
                    self.assertEqual(message, None)
                else:
                    self.assertEqual(message, TestDPM.message)
                intercomm.Free()
        MPI.COMM_WORLD.Barrier()


MVAPICH2 = MPI.get_vendor()[0] == 'MVAPICH2'
try:
    if MVAPICH2: raise NotImplementedError
except NotImplementedError:
    unittest.disable(TestDPM, 'mpi-dpm')


if __name__ == '__main__':
    unittest.main()



cdef Any
cdef Union
cdef Literal
cdef Optional
cdef NoReturn

cdef Callable
cdef Hashable
cdef Iterable
cdef Iterator
cdef Sequence
cdef Mapping

cdef Tuple
cdef List
cdef Dict

cdef Buffer

cdef BufSpec
cdef BufSpecB
cdef BufSpecV
cdef BufSpecW
cdef TargetSpec



from mpi4py import MPI
import mpiunittest as unittest


@unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, 'mpi-world-size<2')
class BaseTestIntercomm(object):

    BASECOMM  = MPI.COMM_NULL
    INTRACOMM = MPI.COMM_NULL
    INTERCOMM = MPI.COMM_NULL

    def setUp(self):
        size = self.BASECOMM.Get_size()
        rank = self.BASECOMM.Get_rank()
        if rank < size // 2 :
            self.COLOR = 0
            self.LOCAL_LEADER = 0
            self.REMOTE_LEADER = size // 2
        else:
            self.COLOR = 1
            self.LOCAL_LEADER = 0
            self.REMOTE_LEADER = 0
        self.INTRACOMM = self.BASECOMM.Split(self.COLOR, key=0)
        Create_intercomm = MPI.Intracomm.Create_intercomm
        self.INTERCOMM = Create_intercomm(self.INTRACOMM,
                                          self.LOCAL_LEADER,
                                          self.BASECOMM,
                                          self.REMOTE_LEADER)

    def tearDown(self):
        self.INTRACOMM.Free()
        self.INTERCOMM.Free()

    def testFortran(self):
        intercomm = self.INTERCOMM
        fint = intercomm.py2f()
        newcomm = MPI.Comm.f2py(fint)
        self.assertEqual(newcomm, intercomm)
        self.assertTrue(type(newcomm) is MPI.Intercomm)

    def testLocalGroupSizeRank(self):
        intercomm = self.INTERCOMM
        local_group = intercomm.Get_group()
        self.assertEqual(local_group.size, intercomm.Get_size())
        self.assertEqual(local_group.size, intercomm.size)
        self.assertEqual(local_group.rank, intercomm.Get_rank())
        self.assertEqual(local_group.rank, intercomm.rank)
        local_group.Free()

    def testRemoteGroupSize(self):
        intercomm = self.INTERCOMM
        remote_group = intercomm.Get_remote_group()
        self.assertEqual(remote_group.size, intercomm.Get_remote_size())
        self.assertEqual(remote_group.size, intercomm.remote_size)
        remote_group.Free()

    def testMerge(self):
        basecomm  = self.BASECOMM
        intercomm = self.INTERCOMM
        if basecomm.rank < basecomm.size // 2:
            high = False
        else:
            high = True
        intracomm = intercomm.Merge(high)
        self.assertEqual(intracomm.size, basecomm.size)
        self.assertEqual(intracomm.rank, basecomm.rank)
        intracomm.Free()


class TestIntercomm(BaseTestIntercomm, unittest.TestCase):
    BASECOMM = MPI.COMM_WORLD

class TestIntercommDup(TestIntercomm):
    def setUp(self):
        self.BASECOMM = self.BASECOMM.Dup()
        super(TestIntercommDup, self).setUp()
    def tearDown(self):
        self.BASECOMM.Free()
        super(TestIntercommDup, self).tearDown()

class TestIntercommDupDup(TestIntercomm):
    def setUp(self):
        super(TestIntercommDupDup, self).setUp()
        INTERCOMM = self.INTERCOMM
        self.INTERCOMM = self.INTERCOMM.Dup()
        INTERCOMM.Free()


if __name__ == '__main__':
    unittest.main()
